(* Copyright (C) 2024     Marc Mezzarobba *)
(* AbsolutePrimeDecompositionBySingular(J) inputs a POLYNOMIALIDEAL structure
   and outputs an exprseq apd, tim, ver consisting of:
   * a pair apd whose first and second entries are:
     * a list of pairs [I, 'conjugates'=n] where I is a _prime_ POLYNOMIALIDEAL
       that potentially involves a RootOf and n is the number of conjugates of
       the prime ideal I;
     * a set of all RootOf involved in the ideals I;
   * an integer tim measuring the time in ms spent in Singular;
   * an integer ver denoting the Singular release, e.g., 4328 for development
     version 4.3.2 with patch 8.
  Notes:
  * Input/output use RootOf, no sqrt.
  * This calls the procedure absPrimdecGTZ of Singular, which has to be
    installed. *)
local AbsolutePrimeDecompositionBySingular := module()

local singular_template, collection_to_string;
export ModuleApply;

singular_template :=
    "timer = 0;\n"
    "system(\"--ticks-per-sec\", 1000);\n"
    "int tim = timer;\n"
    "printlevel = -1;\n"
    "short = 0;\n"
    "LIB \"primdec.lib\";\n"
    "ring  r = 0,(VARS),lp;\n"
    "ideal i = POLYS;\n"
    "def S = absPrimdecGTZ(i);\n"
    "tim = timer - tim;\n"
    "setring S;\n"
    "int ver = system(\"version\");\n"
    "list lst = absolute_primes, tim, ver;\n"
    "print(lst, \"%l\");\n"
    "quit;\n"
    ;

collection_to_string := proc(collection)
    StringTools[Join](map(convert, [op(collection)], 'string'), ",");
end proc;

ModuleApply := proc(J, $)
    local filename, i, out, process_ideal, status;
    local inivars := [op(PolynomialIdeals[IdealInfo][Variables](J))];
    local vars := [seq(cat('x', i), i=1..nops(inivars))];
    local gbasis := Groebner[Basis](J,
            PolynomialIdeals[IdealInfo][DefaultMonomialOrder](J));
    local polys := subs(zip(`=`, inivars, vars), gbasis);
    local singular_input := StringTools[Subs](
            ["VARS" = collection_to_string(vars),
             "POLYS" = collection_to_string(polys)],
            singular_template);
    try
        filename := FileTools[TemporaryFile]();
        FileTools[Text][WriteString](filename, singular_input);
        FileTools[Text][Close](filename);
        status, out := op(ssystem(cat("Singular -q ", filename)));
        if status != 0 then
            error "Singular returned exit status %1, input=%2, output=%3",
                  singular_input, out;
        end if;
        if StringTools[Search]("?", out) > 0 then
            userinfo(1, 'hints', "Singular Input:", print(singular_input));
            userinfo(1, 'hints', "Singular Output:", print(out));
            error "Singular failed to compute the primary decomposition";
        end if;
    finally
        FileTools[Remove](filename);
    end try;
    process_ideal := proc()
        local gens, algeqs;
        gens, algeqs := selectremove(has, [_rest], vars);
        local sol := SolveTools[PolynomialSystem](algeqs);
        if sol = NULL then
            gens := [1]
        else
            gens := subs(zip(`=`, vars, inivars), subs(sol, gens));
        end if;
        PolynomialIdeals[PolynomialIdeal](gens, 'variables'=inivars);
    end proc;
    local apd, tim, ver, Int;
    apd, tim, ver := op(eval(parse(out),
                             ['ideal' = process_ideal,
                              'int' = Int,
                              'list' = proc() [_rest] end]));
    apd := eval(apd, 'int' = proc(x) 'conjugates'=x end);
    tim, ver := op(tim), op(ver);
    return apd, tim, ver
end proc;

end module:
