(* 
                  Copyright (C) 2024 Philippe Dumas
  
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License  version 3
  as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  See <https://www.gnu.org/licenses/gpl-3.0.html> for more details
  about the GNU General Public License version 3.
  
  ---------------------------------------------------------------------------
  Maple package dcfun, version 0.0.0, seal f1a1ee7.

  The dcfun package is dedicated to the study of linear Mahler equations. 
  In an equivalent way, it allows (or rather will allow) 
  to investigate linear recurrences of the Divide-and-Conquer type,
  hence its name.
  
  Help is available on the web page https://mathexp.eu/dumas/dcfun/ .
  
*)

dcfun := module()
option package;
export
  version,
  LMOpEval,
  MOp,
  GOp,
  GROp,
  RamRatPolyToLMOp,
  LMOpMul,
  LMOpRightEuclideanDivision,
  LMOpNormalization,
  LMOpGCRD,
  LMOpLCLM,
  LMOpChange,
  LMOpNewtonPolygon,
  LMOpSolve,
  LMOpTruncation,
  RiccatiSolve,
  GosperPetkovsekForm,
  MHypergeom,
  LMOpConvexHull,
  RamificationOrder,
  PuiseuxSeries;
global
  _C,
 `type/radix`,
 `type/ramratpoly`,
 `type/laurentseries`,
 `type/puiseuxseries`,
 `print/MHypergeom`,
 `print/PuiseuxSeries`;
local
  seal,
  SOp,
  cachetableSOp,
  CommutationSOpMOpRule,
  RatSOp,
  RatSOpByResultant,
  RatSOpBygfun,
  PolySOp,
  CKMRSOp,
  TestTosubsArgSOp,
  cachetableTestTosubsArgSOp,
  LMOpToFF,
  cachetableGOp,
  cachetableGROp,
  RRP2LMOpByLCLM,
  RRP2LMOpByLA,
  PowerFunctionToLMOp,
  NaiveLMOpRightEuclideanDivision,
  LMOpHornerScheme,
  LMOpSection,
  LMOpSplit,
  JustDoLMOpGCRD,
  LMOpLCLMBCLS12,
  LMOpLCLMBostanChyzakLiSalvy2012,
  MFlattening,
  MSylvester,
  LMOpLCLMDumas1993,
  LMOpLCLMD93,
  LMOpToPolynomialSolutions,
  LMOpToRationalFunctionsSolutions,
  LMOpRatSol2018,
  ParamountDenominator,
  LMOpRatSol2022,
  LMOpToLaurentSeriesSolutions,
  LMOpToPuiseuxSeriesSolutions,
  LMOpSupport,
  cachetableLMOpSupport,
  LMOpMonomials,
  cachetableLMOpMonomials,
  RiccatiByPetkovsek,
  ImprovedPetkovsek,
  RiccatiRatSolCleaning,
  BasicPetkovsek,
  AdmissibleEdges,
  RiccatiByHermitePade,
  AbsolutePrimeDecomposition,
  HeurAbsPrimeDecomp,
  HermitePadeApproximants,
  LMOpToMHypergeometricSolutions,
  EvanescentProductCleaning,
  cachetableEvanescentProductCleaning,
  MSFToSiblingsGroups,
  IQuo,
  cachetableLMOpConvexHull,
  JustDoLMOpConvexHull,
  CharPolysConvexHull,
  GeomConvexHull,
  Cross,
  ExtendedInverse,
  Slope,
  Intercept,
  SeriesValuation,
  NextTuple,
  RatPolyToMSF,
  MSFToMSF,
  MSFRatPolyMultiply,
  MSFPolyDivide,
  WellChosenDenominator,
  JustDoRamificationOrder,
  CommonRamificationOrder,
  CompRatFunSets,
  ComparingRationalFunctions,
  RationalFunctionsOrientedComparison,
  RenamingParameters,
  FieldExtension1,
  FieldExtension2;
local
  _version := "0.0.0";
seal := "f1a1ee7";
unprotect(PuiseuxSeries):
protect(_C):
unprotect(version):
#
# current version
version := proc() 
  return _version 
end proc:# version
#
SOp := proc(expr::algebraic, x::name, b::{posint,name}, r)
  local
    residues_number, k, msg, result, val, deg, kmin, kmax, prec,
    q1, i, expr2, expr1, s, h, content, b1, r1, fct, x1, g, rbar,
    bbar, b1bar, k0, op1expr, D, c,
    n0, remain, u, v;


  if _npassed > 4 then
    return CKMRSOp(_passed)
  fi;
  
  if type(r, 'list') then
    userinfo(3,'procname',"list case");
    residues_number := nops(r);
    if residues_number > 1 then
      return SOp(SOp(expr,x,b,r[residues_number]),
                  x,b,[seq(r[k],k=1..residues_number-1)])
    elif residues_number = 1 then
      return SOp(expr,x,b,r[1])
    else
      return expr
    fi
  fi;
  
  if type(b, 'integer') then
    if b < 1 then
      msg := "the first argument is expected to be an integer > 0, ";
      msg := cat(msg,"but %1 is received");
      error msg, b
    fi;
  else
    return 'procname(_passed)'
  fi;
  if type(r, 'integer') then
    if (r < 0 or r >= b) then
    msg := "the fourth argument is assumed to be an integer ";
    msg := cat(msg,"between 0 and %1, ");
    msg := cat(msg,"but %2 is received");
      error msg, b-1, r
    fi
  else
     return 'procname(_passed)'
  fi;
  
  if b = 1 and r = 0 then 
    return expr
  fi;
  
  if type(expr, 'polynom'(anything,x)) then
    userinfo(3,'procname',"polynomial case");
    result := PolySOp(_passed);
    return result
  fi;
  if type(expr, 'ratpoly') then
    userinfo(3,'procname',"rational function case");
    if evalb(expr = 1/(1-x)) then
      return 1/(1-x)
    fi;
    result := RatSOp(_passed);
    return result
  fi;
  
  if type(expr, 'series') then
    userinfo(3,'procname',"series case");
    result := convert(expr,polynom);
    val := ldegree(result,x);
    deg := degree(result,x);
    if val < 0 then
      kmin := -1 + iquo(val,b)
    else
      kmin :=    + iquo(val,b)
    fi;
    kmax := 1 + iquo(deg,b);
    result := add(coeff(result,x,b*k+r)*x^k,k=kmin..kmax);
    if op(-2,expr) = O(1) then
      prec := op(-1,expr);
      q1 := IQuo(prec,b,'r1');
      if r < r1 then q1 := q1 + 1 fi;
      return series(result + O(x^q1),x,q1)
    else
      return series(result,x,infinity)
    fi;
  fi;
  if type(expr, `+`) then
    userinfo(3,'procname',"sum case");
    return add(SOp(op(i,expr),x,b,r),i = 1..nops(expr))
  fi;
  if type(expr, `*`) then
    userinfo(3,'procname',"product case");
    expr2 := op(-1,expr);
    expr1 := normal(expr/expr2);
    return add(x^(iquo(s + (r-s mod b),b))*
        SOp(expr1,x,b,s)*SOp(expr2,x,b,r-s mod b),s=0..b-1)
  fi;
  if type(expr, 'function') then
    userinfo(3,'procname',"function case");
    h := op(0,expr);
    content := [op(expr)];
    if h = 'SOp' then
      b1 := content[3];
      r1 := content[4];
      fct := content[1];
      x1 := content[2];
      if x1 <> x then
        msg := "we are dealing with univariate functions, ";
        msg := cat(msg,"but you are using two variables %1 and %2");
        error msg, x, x1
      fi;
      return 'SOp'(fct,x,b*b1,b1*r+r1)
    elif h = 'MOp' then
      b1 := content[3];
      fct := content[1];
      x1 := content[2];
      if x1 <> x then
        msg := "we are dealing with univariate functions, ";
        msg := cat(msg,"but you are using two variables %1 and %2");
        error msg, x, x1
      fi;
      g := igcd(b,b1);
      rbar := iquo(r,g,'remain');
      if remain = 0 then
        bbar := b/g;
        b1bar := b1/g;
        igcdex(bbar,b1bar,'u','v');
        IQuo(v*rbar,bbar,'n0');
        k0 := IQuo(-rbar + b1bar*n0,bbar);
        return x^k0*'MOp'(SOp(fct,x,bbar,n0),x,b1bar)
      else
        return 0
      fi
    else
      if type(op(1,expr), 'monomial') then
        userinfo(3,'procname',"monomial inside");
        op1expr := op(1,expr);
        if op1expr = x then return 'SOp'(h(x),x,b,r) fi;
        D := degree(op1expr,x);
        c := coeff(op1expr,x,D);
        if c = 1 then
          return CommutationSOpMOpRule(h(x),x,b,r,D)
        else
          msg := cat("perhaps it would be a good idea to change ",
            convert(x,string)," into ",convert(c^(1/D)*x,string));
          WARNING(msg);
          return 'SOp'(_passed)
        fi
      fi;
    fi
  fi;
  
  return 'SOp'(_passed)
end proc:# SOp
cachetableSOp := Cache(100):
Cache(cachetableSOp, procedure=SOp):
#
CommutationSOpMOpRule := proc(f, x, b, r, b1)
  local g, rbar, bbar, b1bar, k0, remain, u, v, n0;
  option procname;
  g := igcd(b, b1);
  rbar := iquo(r, g, 'remain');
  if remain = 0 then
    bbar := b/g;
    b1bar := b1/g;
    igcdex(bbar, b1bar, 'u', 'v');
    IQuo(v*rbar, bbar, 'n0');
    k0 := IQuo(-rbar + b1bar*n0, bbar);
    return x^k0*MOp(SOp(f, x, bbar, n0), x, b1bar)
  else
    return 0
  fi
end proc:# CommutationSOpMOpRule
#
RatSOp := proc(f::ratpoly(anything,x), x::name, b::posint, r,
  {method::Or(identical(resultant),identical(hermite_pade)):='resultant'})
  local msg;
  option procname;
  if (r < 0 or r >= b) or not type(r, integer) then
  msg := "the fourth argument is assumed to be an integer ";
  msg := cat(msg, "between 0 (included) and %1 (excluded), ");
  msg := cat(msg, "but %2 is received");
    error msg, b, r
  fi;
  if method = 'resultant' then
    return RatSOpByResultant(f, x, b, r)
  else
    return RatSOpBygfun(f, x, b, r)
  fi
end proc:# RatSOp
#
RatSOpByResultant := proc(f,x,b,r)
  option procname;
  local ff := normal(f);
  local nn := numer(ff);
  local dd := denom(ff);
  local _Z;
  local rr := resultant(_Z^b - x^b, subs(x = '_Z', dd), '_Z');
  local qq;
  divide(rr, dd, 'qq');
  return normal(PolySOp(expand(nn * qq), x, b, r)
                  /  PolySOp(rr, x, b, 0))
end proc:# RatSOpByResultant
#
RatSOpBygfun := proc(f,x,b,r)
  option procname;
  local ff := normal(f);
  if ff = 0 then return 0 fi;
  local nn := numer(ff);
  local dd := denom(ff);
  local nu := degree(nn,x);
  local delta := degree(dd,x);
  local deq := dd * '_Z'(x) - nn;
  local rec := gfun:-diffeqtorec(deq,'_Z'(x),'u'('n'));
  local uproc := gfun:-rectoproc(rec,'u'('n'),remember);
  return convert(series(add(uproc(b*nn+r)*x^nn,nn=0..nu+delta+1),
                        x,nu+delta+1),ratpoly,nu,delta);
end proc:# RatSOpBygfun
#
PolySOp := proc(p,x,b,r)
  option procname;
  local expanded_p := (expand@normal)(p);
  local deg := degree(expanded_p,x);
  local val := ldegree(expanded_p,x);
  local k;
  return add(coeff(expanded_p,x,b*k+r)*x^k,k=iquo(val,b)..iquo(deg,b));
end proc:# PolySOp
#
CKMRSOp := proc(f, x, b, r, tosubs::set(`=`(function, algebraic)))
local newtosubs, msg, res;
  option procname;
  newtosubs := TestTosubsArgSOp(x, b, tosubs);
  if newtosubs = FAIL then
    msg := cat("It is expected that the last argument tosubs is a set of ",
      "equations u(%1) = a linear homogeneous combination of some ",
      "v(%1^(%2^k)) with k a non negative integer and u, v in a finite set U ",
      "of names, and also that this system defines univoquely the u(%1) ",
      "as functions of the v(%1^(%2^k)) with k a positive integer, but ",
      "%3 is received");
      error msg, x, b, tosubs
  fi;
  res := SOp(eval(f, newtosubs), x, b, r);
  res := collect(res, map(z->op(1,z), newtosubs), normal);
  return res
end proc:# CKMRSOp
#
TestTosubsArgSOp := proc(x, b, tosubs)
  local fun1, uset, uofx, u, xx, tosubsrhs, fun2, opfun2, logopfun2,
    tsbs, g, formalfun2, formaltosubsrhs, ff, h,
    newtosubs, newfun1, newfun2, z, homgenpart, nonhomgenpart;
  option procname;
  fun1 := map(z-> op(1,z), tosubs);
  uset := {};
  for uofx in fun1 do
    u := op(0, uofx);
    uset := uset union {u};
    xx := op(1, uofx);
    if not (xx = x and nops(uofx) = 1) then return FAIL fi;
  od;
  tosubsrhs := map(z->op(2, z), tosubs);
  fun2 := indets(tosubsrhs, function);
  fun2 := select(has, fun2, uset);
  if not member(map(nops, fun2), {{},{1}}) then
    return FAIL
  fi;
  opfun2 := map(op, fun2);
  if not (type(opfun2, 'set'('monomial'))
          and indets(opfun2) subset {x}
          and member(map2(remove, has, opfun2, x), {{},  {'undefined'}}))
  then return FAIL
  fi;
  logopfun2 := map(z->ln(op(2,z))/ln(b), opfun2 minus {x}); 
  if not type(logopfun2, set(posint)) then
    return FAIL
  fi;
  tsbs := {seq(g = h[g], g = fun2)};
  formalfun2 := eval(fun2, tsbs);
  formaltosubsrhs := eval(tosubsrhs, tsbs);
  for ff in formaltosubsrhs do 
    nonhomgenpart := normal(eval(ff, {seq(h = 0, h = formalfun2)}));
    homgenpart := normal(ff - nonhomgenpart);
    if not ( (homgenpart = 0) or
            (
              type(homgenpart, polynom(ratpoly(anything, x),formalfun2))
              and degree(homgenpart, formalfun2) = 1 
              
            )
           ) then
      return FAIL
    fi
  od;

  if has(tosubsrhs, fun1) then
    try newtosubs := solve(tosubs, fun1)
    catch :
      return FAIL
    end try
  else
    newtosubs := tosubs
  fi;
  newtosubs := map(z-> op(1,z)
                    = collect(op(2,z), indets(newtosubs, function), normal),
                  newtosubs);
  newfun1 := map(z-> op(1,z), newtosubs);
  newfun2 := map(z-> op(2,z), newtosubs);
  if newfun1 <> fun1 or has(newfun2, newfun1) then
    return FAIL
  fi;
  return newtosubs
end proc:# TestTosubsArgSOp
cachetableTestTosubsArgSOp := Cache(100):
Cache(cachetableTestTosubsArgSOp, procedure=TestTosubsArgSOp):
#
# apply a linear Mahler operator to an expression
LMOpEval := proc(L::depends(polynom(ramratpoly(anything,x), M)),
                  f::algebraic,
                  x::name, M::name, b::radix, $)
  local Q, LL, ff, DENOM, r0, r, k, ell, alpha::nothing, omega, delta,
    Alpha, Omega, Alpha_plus, res, q, s, kk, g::nothing, tosubs::nothing;
  Q := RamificationOrder(L, x);
  if Q > 1 then
    LL := MOp(L, x, Q); ff := MOp(f, x, Q);
    res := normal(LMOpEval(LL, ff, x, M, b));
    return MOp(res, x, 1/Q)
  fi;

  LL, DENOM := LMOpToFF(L, M);
  r0 := ldegree(LL, M);
  r := degree(LL, M);
  for k from r0 to r do
    ell[k] := expand(coeff(LL, M, k))
  od;

  if patmatch(f, (ln(x)^('alpha'::algebraic)
                      * ('g'::algebraic)), 'tosubs') then
    eval(ln(x)^alpha, tosubs)
    * LMOpEval(eval(L, M = simplify(eval(b^alpha, tosubs), 'power')*M),
               eval(g, tosubs), x, M, b)
  elif patmatch(f, (ln(1/x)^('alpha'::algebraic)
                          * ('g'::algebraic)), 'tosubs') then
    eval(ln(1/x)^alpha, tosubs)
    * LMOpEval(eval(L, M = simplify(eval(b^alpha, tosubs), 'power')*M),
               eval(g, tosubs), x, M, b)
  elif type(f, 'laurentseries'(anything, x)) and op(0, f) = x then
    DENOM := expand(DENOM);
    for k from r0 to r do
      delta[k] := ldegree(ell[k], x)
    od;
    alpha := op(2, f);
    omega := op(-1, f);
    
    
    if op(-2, f) <> O(1) then
      return LMOpEval(L,
              series(f + O(x^(omega +1)), x), x, M, b)
    fi;

    Alpha := min(seq(delta[k] + b^k*alpha, k=r0..r));
    Omega := min(seq(delta[k] + b^k*omega, k=r0..r));
    Alpha_plus :=`if`(Alpha >= 0, 0, -Alpha);
    Order := Alpha_plus + Omega + max(seq(delta[k], k = r0..r));
    for k from r0 to r do
      res[k] := MOp(f, x, b^k);
      res[k] := series(ell[k]*res[k], x); 
    od;
    res := series(add(res[k], k = r0..r), x);
    return map(evala@normal, series(res/DENOM, x))
  elif type(f, 'puiseuxseries'(anything, x)) then
    q := op(1, f);
    s := op(2, f);
    return 'PuiseuxSeries'(q, LMOpEval(eval(L, x = x^q), s, x, M, b))
  elif type(f, `+`) then
    add(LMOpEval(L, g, x, M, b), g = f)
  else
    res := add(ell[kk]*MOp(f, x, b^kk), kk = r0..r);
    return normal(res/DENOM)
  fi
end proc:# LMOpEval
#
LMOpToFF := proc(L, M)
  option procname;
  local LL;
  LL := Normalizer(L);
  return collect(numer(LL), M, expand), denom(LL)
end proc:# LMOpToFF
#
# the Mahler operator
MOp := proc(f, x::name, b, $)
  local h, content, msg, fct, x1, b1, alpha::nothing, tosubs::nothing, bb, ell, i;
  if not (type(b,rational) and b > 0) then
    return 'procname(_passed)'
  fi;
  if not has(indets(f, name), x) then 
    return f
  fi;
  if b = 1 then return f fi;
  if type(f, `+`) then
    userinfo(3,'procname',"+ case");
    return map(MOp,f,x,b)
  fi;
  if type(f, `*`) then
    userinfo(3,'procname',"* case");
    return map(MOp,f,x,b)
  fi;
  if patmatch(f, (x^('alpha'::constant)), 'tosubs') then
    return simplify(eval(x^(b*alpha), tosubs), 'power', 'symbolic')
  fi;
  if patmatch(f, (ln(x)^('alpha'::constant)), 'tosubs') then
    return simplify(eval(b^alpha, tosubs), 'power', 'symbolic')
             * eval(ln(x)^alpha, tosubs)
  fi;
  if patmatch(f, (ln(1/x)^('alpha'::constant)), 'tosubs') then
    return simplify(eval(b^alpha, tosubs), 'power', 'symbolic')
             * eval(ln(1/x)^alpha, tosubs)
  fi;

  if type(f, 'function') then
    userinfo(3,'procname',"function case");
    h := op(0,f);
    content := [op(f)];
    msg := "we are dealing with univariate functions, ";
    msg := cat(msg,"but you are using two variables %1 and %2");
    if h = 'MOp' then
      userinfo(3,'procname',"Mahler operator case");
      fct := content[1];
      x1 := content[2];
      b1 := content[3];
      if x1 <> x then
        error msg, x, x1
      fi;
      return 'MOp'(fct,x,b*b1)
    fi;
    if h = 'SOp' then
      userinfo(3,'procname',"Section operator case");
      fct := content[1];
      x1 := content[2];
      b1 := content[3];
      if x1 <> x then
        error msg, x, x1
      fi;
      return 'MOp'(_passed)
    fi;
    if h = 'MHypergeom' then
      if nops(content) = 3 and content[2] = x then
        bb := content[3];
        ell := log[bb](b);
        if type(ell, 'nonnegint') then
          return product(MOp(content[1], x, bb^i), i = 0..ell - 1) * f
        elif type(ell, 'negint') then
          return 1/product(MOp(content[1], x, 1/bb^i), i = 1..-ell) * f
        else
          return MHypergeom(MOp(content[1], x, b), x, bb)
        fi
      else
        error "badly defined Mahler hypergeometric function: %1", f
      fi
    fi;
    if h = 'PuiseuxSeries' then
      return PuiseuxSeries(op(1, f), MOp(op(2, f), x, b))
    fi;
    if nops(content) = 1 then
      userinfo(3,'procname',"function case");
      return simplify(eval(f,x=x^b), 'power', 'symbolic')
    fi;
    userinfo(3,'procname',"generic case");
    return 'MOp'(_passed)
  elif type(f, 'series') then
    userinfo(3,'procname',"series case");
    return eval(f,x=x^b)
  elif type(f, 'algebraic') then
    userinfo(3,'procname',"algebraic case");
    return simplify(eval(f,x=x^b), 'power', 'symbolic')
  else
    userinfo(3,'procname',"generic case");
    'MOp'(_passed)
  fi
end proc:# MOp
#
# Gräffe operator
GOp := proc(p, x::name, b::{posint,name}, $)
  if type(b, name) then
    return 'procname'(_passed)
  fi;
  if type(p, polynom(anything, x)) then
    return collect(Algebraic:-Resultant('_Z'^b - x, eval(p, x='_Z'), '_Z'), x)
  else
    return 'procname'(_passed)
  fi;
end proc:# GOp
cachetableGOp := Cache(100):
Cache(cachetableGOp, procedure=GOp):
#
# radical of the Gräffe operator
GROp := proc(p, x::name, b::{posint,name}, $)
  local msg, g;
  if type(b, name) then
    return 'procname'(_passed)
  fi;
  if type(p, polynom(anything, x)) then
    g := GOp(p, x, b);
    return normal(g/gcd(g, diff(g, x)))
  else
    return 'procname'(_passed)
  fi;
end proc:# GROp
cachetableGROp := Cache(100):
Cache(cachetableGROp, procedure=GROp):
#
# from a ramified rational function to a linear Mahler operator
RamRatPolyToLMOp := proc(f::ramratpoly(anything,x),
  x::{name,symbol}, M::name, b::radix,
  {method::Or(identical(RRP2LMOpByLCLM),identical(RRP2LMOpByLA))
    :='RRP2LMOpByLA',
   output::Or(identical(polynomial),identical(monic)):='polynomial'}, $)
  local Mf_over_f;
  if normal(simplify(f, 'power', 'symbolic')) = 0 then return 1 fi;
  if type(f, 'ratpoly'(anything, x)) then
    Mf_over_f := normal(MOp(f, x, b)/f);
    if 'output' = 'polynomial' then
      return numer(Mf_over_f) - denom(Mf_over_f)*M
    else
      return M - Mf_over_f
    fi
  fi;
  return method(f, x, M, b, _options['output'])
end proc:# RamRatPolyToLMOp
#
RRP2LMOpByLCLM := proc(f, x, M, b,
  {output::Or(identical(polynomial),identical(monic)):='polynomial'}, $)
  local q, f1, f1_tb, L_tb, t, L;
  q := RamificationOrder(f, x);
  f1 := simplify(eval(f, x = x^q), 'power', 'symbolic');
  f1_tb := table();
  L_tb := table();
  f1_tb[0] := normal(SOp(f1, x, q, 0));
  L_tb[0] := PowerFunctionToLMOp(x^(0/q) * f1_tb[0], x, M, b);
  for t from 1 to q - 1 do
    f1_tb[t] := normal(SOp(f1, x, q, t));
    L_tb[t] := PowerFunctionToLMOp(x^(t/q) * f1_tb[t], x, M, b);
  od;
  L := LMOpLCLM([seq(L_tb[t], t = 0..q-1)], x, M, b);
  if 'output' = 'polynomial' then
    L := (numer@normal)(L);
    L := collect(L, M, normal)
  else
    L := collect(L/lcoeff(L, M), M, normal)
  fi;
  return L
end proc:# RRP2LMOpByLCLM
#
RRP2LMOpByLA := proc(f, x, M, b,
  {output::Or(identical(polynomial),identical(monic)):='polynomial'}, $)
local q, q1, Bq1, k, omega, g, u, Sigma, L1, r, ell, sys, inc, sol, ELL, L;
  q := RamificationOrder(f, x);
  q1 := q; Bq1 := 1/q;
  for k while gcd(b, q1) <> 1 do
    Bq1 := b*Bq1; q1 := denom(Bq1)
  od;
  omega := k - 1;
  g := simplify(MOp(f, x, b^omega*q1), 'power', 'symbolic');
  for u from 0 to q1 - 1 do
    Sigma[u] := SOp(g, x, q1, u)
  od;
  L1 := 0;
  for r while L1 = 0 do
    sys := {seq(add(ell[k] *
                    x^floor(b^k * (b^(-k)*u mod q1)/q1) *
                    MOp(Sigma[b^(-k)*u mod q1], x, b^k), k = 0..r),
                u = 0..q1 - 1)};
    inc := {seq(ell[k], k = 0..r)};
    sol := solve(sys, inc);
    L1 := normal(eval(add(ell[k]*M^k, k = 0..r), sol))
  od;
  ELL := op(select(has, indets(L1, name), ell));
  L1 := normal(L1/ELL*M^omega);
  if 'output' = 'polynomial' then
    L := collect(numer(L1), M)
  else
    L1 := collect(L1, M, normal);
    L := collect(L1/lcoeff(L1, M), M, normal)
  fi;
  return L
end proc:# RRP2LMOpByLA
#
PowerFunctionToLMOp := proc(f::algebraic, x::name, M::name, b::posint)
  local  msg, alpha, Omega, p, q, k, g, beta, L, tosubs::nothing,
    la::nothing, al::nothing, ratio::nothing, be::nothing, rat, LL, ff, r;
  option procname;
  if b < 2 then error "an integer > 1 is expected as 4th argument" fi;
  if normal(simplify(f, 'power', 'symbolic')) = 0 then return 1 fi;
  if type(f, 'ratpoly'(anything, x)) then
    return RamRatPolyToLMOp(f, x, M, b)
  fi;
  msg := "there is no linear Mahler operator ";
  msg := cat(msg,"for the radix %1 whose coefficients are ");
  msg := cat(msg,"rational functions and which annihilates %2");

  if patmatch(f, ('la'::constant)*x^('al'::constant)
                *('ratio'::ratpoly(anything, x)),'tosubs') then
    alpha := eval(al,tosubs);
    rat := eval(ratio, tosubs);
    p := numer(alpha);
    q := denom(alpha);
    if q = 1 then
      k := 1
    else
      g := igcd(b,q);
      if g<> 1 then
        Omega := 0;
        while g <> 1 do
          alpha := p*b/q;
          p := numer(alpha);
          q := denom(alpha);
          Omega := Omega + 1;
          g := igcd(b,q);
        od;
        return PowerFunctionToLMOp(
          x^(p/q) * MOp(rat, x, b^Omega), x, M, b) * M^Omega
      else
        k := NumberTheory:-MultiplicativeOrder(b,q);
      fi
    fi;
    if p > 0 then return
      (numer@normal)(M^k - x^(alpha*(b^k-1))*MOp(rat, x, b^k)/rat)
    else return
     (numer@normal)(-MOp(rat, x, b^k)/rat + x^(-alpha*(b^k-1)) * M^k)
    fi
  fi;
  if patmatch(f, ('la'::constant)
                    *ln(x)^('be'::constant)
                  *('ratio'::ramratpoly(anything, x)),'tosubs') then
    beta := eval(be,tosubs);
    L := RamRatPolyToLMOp(eval(la*ratio,tosubs), x, M, b);
    r := degree(L, M);
    return collect(b^(r*beta)*eval(L, M = M/b^beta), M, `simplify/power`)
  fi;
  if patmatch(f, ('la'::constant)
                    *ln(1/x)^('be'::constant)
                  *('ratio'::ramratpoly(anything, x)),'tosubs') then
    beta := eval(be,tosubs);
    L := RamRatPolyToLMOp(eval(la*ratio,tosubs), x, M, b);
    r := degree(L, M);
    return collect(b^(r*beta)*eval(L, M = M/b^beta), M, `simplify/power`)
  fi;
  if patmatch(f, ('la'::constant)*x^('al'::constant)
                  *ln(x)^('be'::constant)
                *('ratio'::ratpoly(anything, x)),'tosubs') then
    beta := eval(be,tosubs);
    try
      L := PowerFunctionToLMOp(eval(la*x^al*ratio,tosubs), x, M, b)
    catch :
      error msg, b, f
    end try;
    r := degree(L, M);
    return collect(b^(r*beta)*eval(L, M = M/b^beta), M, `simplify/power`)
  fi;
  if patmatch(f, ('la'::constant)*x^('al'::constant)
                  *ln(1/x)^('be'::constant)
                *('ratio'::ratpoly(anything, x)),'tosubs') then
    beta := eval(be,tosubs);
    try
      L := PowerFunctionToLMOp(eval(la*x^al*ratio,tosubs), x, M, b)
    catch :
      error msg, b, f
    end try;
    r := degree(L, M);
    return collect(b^(r*beta)*eval(L, M = M/b^beta), M, `simplify/power`)
  fi;
  if type(f, `+`) then
    try
      LL := table();
      for  ff in f do
        LL[ff] := PowerFunctionToLMOp(ff, x, M, b)
      od;
    catch :
      error msg, b, f
    end try;
    L := LMOpLCLM([seq(LL[ff], ff = f)], x, M, b);
    return L
  fi;
  msg := "the first argument %1 does not match ";
  msg := cat(msg,"the specification of the procedure");
  error msg, f
end proc:# PowerFunctionToLMOp
#
# multiplication of linear Mahler operators
LMOpMul := proc(L1::depends(polynom(anything,M)),
    L2::depends(polynom(anything,M)), x::symbol, M::symbol, b::radix, $)
  local LL1, LL2, d1, d2, d, A, B, i, msg, k, res;

  LL1 := collect(L1,M,normal);
  LL2 := collect(L2,M,normal);
  d1 := degree(LL1,M);
  d2 := degree(LL2,M);
  d := max(d1,d2);

  A := table();
  B := table();
  for i from 0 to d do
    A[i] := coeff(LL1,M,i);
    B[i] := coeff(LL2,M,i)
  od;
  for k from 0 to d do
    res[k] := add(A[i]*MOp(B[k-i], x, b^(i)), i=0..k)
  od;
  for k from d + 1 to d1 + d2 do
    res[k] :=
      add(A[i]*MOp(B[k-i], x, b^(i)), i=k-d..d)
  od;

  return add(normal(res[i])*M^i,i=0..d1+d2)
end proc:# LMOpMul
#
# Euclidean division of linear Mahler operators
LMOpRightEuclideanDivision := proc(
    L1::depends(polynom(anything,M)),
    L2::depends(polynom(anything,M)),
                          x::symbol,M::symbol,b::radix, $)
  local LL1, LL2, Q, d2, c2, test, q, R, msg;

  LL1 := collect(L1,M);
  LL2 := collect(L2,M);
  if type(LL1, 'polynom'('ratpoly'('anything',x),M))
    and type(LL2, 'polynom'('ratpoly'('anything',x),M)) then
    d2 := degree(LL2, M);
    if d2 < 10 then
      c2 := coeff(LL2, M, d2);
      return NaiveLMOpRightEuclideanDivision(LL1, LL2, x, M, b, d2, c2)
    else
      return "not yet implemented"
    fi
  fi;
  q := RamificationOrder({coeffs(LL1, M), coeffs(LL2, M)}, x);
  if q <> FAIL then
    LL1 := MOp(LL1, x, q);
    LL2 := MOp(LL2, x, q);
    Q, R := LMOpRightEuclideanDivision(LL1, LL2, x, M, b);
    Q := collect(MOp(Q, x, 1/q), M);
    R := collect(MOp(R, x, 1/q), M);
    return Q, R
  fi;
  msg := cat("It is expected that LMOpRightEuclideanDivision ",
            "receives two polynomials in %1",
            " of (possibly ramified) rational functions ",
            "in %4, ",
            "but %2 and %3 are received as first and second ",
            "arguments.");
  error msg, M, L1, L2, x
end proc:# LMOpRightEuclideanDivision
#
NaiveLMOpRightEuclideanDivision := proc(L1,L2,x,M,b,d2,c2)
  local d1, c1, q, r, toto;
  option procname;
  d1 := degree(L1, M);
  if d1 < d2 then
    return 0, L1
  else
    c1 := coeff(L1,M,d1);
    q := c1/eval(c2,x=x^(b^(d1-d2))) * M^(d1-d2);
    r := collect(L1 - LMOpMul(q,L2,x,M,b), M,normal);
    toto :=  NaiveLMOpRightEuclideanDivision(r,L2,x,M,b,d2,c2);
    return collect(evala(q + toto[1]), M), collect(evala(toto[2]), M)
  fi
end proc:# NaiveLMOpRightEuclideanDivision
#
LMOpHornerScheme := proc(L::depends(polynom(ratpoly(anything, x), M)),
  u, x::{name, symbol}, M::{name, symbol}, b::posint)
  local LL, r, Q, k, q, newL, newu, R, msg;
  LL := collect(L, M);
  r := degree(LL, M);
  if r = 0 then return LL fi;
  if r = 1 then return eval(LL, M = u) fi;
  if type(u, 'ratpoly'('anything', x)) then
    Q := table();
    Q[r - 1] := normal(coeff(LL, M, r));
    for k from r - 1 by - 1 to 1 do
      Q[k - 1] := normal(coeff(LL, M, k) + Q[k] * MOp(u, x, b^k));
    od;
    return normal(coeff(LL, M, 0) + Q[0] * u)
  fi;
  q := RamificationOrder(u, x);
  if q <> FAIL then
    newL := eval(L, x = x^q);
    newu := simplify(eval(u, x = x^q), 'power', 'symbolic');
    R := LMOpHornerScheme(newL, newu, x, M, b);
    return simplify(eval(R, x = x^(1/q)), 'power', 'symbolic')
  fi;
  msg := "the second argument %1 does not match ";
  msg := cat(msg,"the specification of the procedure");
  error msg, u
end proc:# LMOpHornerScheme
#
LMOpSection := proc(L, x, M, b, r)
  local P, omega, msg, k;
  option procname;
  P := collect(L, M);
  omega := ldegree(P, M);
  if omega = 0 then
    msg := cat("LMOpSection only applies to skew polynomials ",
                "with positive valuation");
    msg := cat(msg,convert(P, string) ,", but has valuation 0");
    error msg
  fi;
  add(SOp(coeff(P,M,k),x,b,r)*M^(k-1), k=omega..degree(P,M))
end proc:# LMOpSection
#
LMOpSplit := proc(L, x, M, b)
  local LL,v,t;
  option procname;
  LL := collect(L,M,expand);
  if LL = 0 then return {} fi;
  v := ldegree(LL,M);
  if v = 0 then return {LL} fi;
  return
    `union`(seq(LMOpSplit(
        LMOpSection(LL,x,M,b,t),x,M,b),t=0..b-1))
end proc:# LMOpSplit
#
# `normalization' of a linear Mahler operator
LMOpNormalization := proc(LL, x::name, M::name, b::radix, $)
  local G;
  G := LMOpSplit(LL, x, M, b);
  return JustDoLMOpGCRD(G, x, M, b);
end proc:# LMOpNormalization
#
# greatest common right divisor for linear Mahler operators
LMOpGCRD := proc(S::{list(polynom(ratpoly(anything,x),M)),
                        set(polynom(ratpoly(anything,x),M))},
                      x::name, M::name, b::radix, $)
  local G, w, g, w1;

  G := convert(S, set);
  G := map(numer@normal, G);
  G := map(collect, G, M, expand);
  w := +infinity;
  for g in G do
    w1 := ldegree(g, M);
    if w1 = 0 then
      w := 0;
      break
    fi;
    if w1 < w then w := w1 fi;
  od;

  G := `union`((seq(LMOpSplit(normal(g/M^w), x, M, b), g=G)));

  return LMOpMul(JustDoLMOpGCRD(G, x, M, b), M^w, x, M, b);
end proc:# LMOpGCRD
#
JustDoLMOpGCRD := proc(S,x,M,b)
  local G, r, rmax, ii, i, F, H, f, h, D;
  option procname;
  G := S;
  while nops(G) > 1 do
    rmax := -1;
    ii := 0;
    for i to nops(G) do
      r := degree(op(i, G), M);
      if r > rmax then
        rmax := r;
        ii := i
      fi
    od;
    F := op(ii, G);
    if ii = 1 then
      H := op(2, G)
    else
      H := op(1, G)
    fi;
    f := coeff(F, M, 0);
    h := coeff(H, M, 0);
    D := collect(f*H - h*F, M, expand);
    G := (G minus {F}) union LMOpSplit(D, x, M, b)
  od;
  return primpart(op(G), M)
end proc:# JustDoLMOpGCRD
#
# lowest common left multiple for linear Mahler operators
LMOpLCLM := proc(Lset::depends({list(polynom(ratpoly(anything,x),M)),
                      set(polynom(ratpoly(anything,x),M))}),
              x::symbol, M::symbol, b::radix,
              cof::name:='undefined',
              {method::Or(identical(LMOpLCLMDumas1993),
                          identical(LMOpLCLMBostanChyzakLiSalvy2012)):=
                          'LMOpLCLMBostanChyzakLiSalvy2012',
   output::Or(identical(polynomial),identical(monic)):='polynomial'})
  local Llist, k, kk, L, D, con::nothing, res, msg, LC;

  Llist := convert(Lset, list);
  k := nops(Llist);
  if k = 0 then return 1 fi;
  if _params[5] <> 'undefined' and not type(_params[5], name) then
    msg := cat("it is expected that the optional fifth argument",
               " be a name, but %1 is received");
    error msg, _params[5]
  fi;
  for kk to k do
    L[kk] := normal(Llist[kk]);
    D[kk] := denom(L[kk]);
    L[kk] := numer(L[kk]);
    L[kk] := primpart(L[kk], M, con[kk]); 
  od;

  if k = 1 then
    if 'output' = 'polynomial' then
      res := L[1]
    else
      res := collect(L[1], M, normal);
      LC := lcoeff(res, M);
      res := collect(res/LC, M, normal)
    fi;
    if _nparams > 4 and _params[5] <> 'undefined' then
      assign(_params[5], D[1]/con[1]/LC)
    fi;
    return res
  fi;

  if _params[5] <> 'undefined' then
    if not type(Lset, list) then
      msg := cat("the first argument must be a list and not a set ",
                 "in order that cofactors be computed, "
                  "but %1 is received.");
      error msg, Lset
    fi
  fi;
  res := method(Llist, _params[2..-1]);
  if 'output' = 'polynomial' then
    LC := 1
  else
    res := collect(res, M, normal);
    LC := lcoeff(res, M);
    res := collect(res/LC, M, normal)
  fi;
  if _params[5] <> 'undefined' then
    cof := eval(_params[5]);
    assign(_params[5],
      [seq(collect(normal(D[kk]/con[kk]*cof[kk]/LC), M, normal),
        kk = 1..k)])
  fi;
  return res

end proc:# LMOpLCLM
#
LMOpLCLMBCLS12 := proc(Llist, x, M, b, cof::name:='undefined')
  local k, kk, LL, rr, s, S, ID,
    Sigma, rho, ell, K, m, L, i, D, c, a, j, U;
  k := nops(Llist);
  if k = 1 then return Llist[1] fi; 
  for kk to k do
    LL[kk] := collect(Llist[kk], M);
    rr[kk] := degree(LL[kk], M);
  od;
  s := add(rr[kk], kk = 1..k);
  for kk to k do
    S[kk] := MSylvester(LL[kk], s, x, M, b);
  od;
  ID := MSylvester(-1, s, x, M, b);
  Sigma := LinearAlgebra:-DiagonalMatrix([seq(S[kk], kk = 1..k)]);
  Sigma := <Sigma ; `<|>`(ID$k)>;
  rho := LinearAlgebra:-Rank(Sigma);
  if assigned('infolevel'[procname]) and 'infolevel'[procname] > 2 then
    printf("matrix Sigma has size %d x %d and rank %d\n",
      LinearAlgebra:-Dimension(Sigma), rho)
  fi;
  ell := rho + s - k * (s + 1);
  if assigned('infolevel'[procname]) and 'infolevel'[procname] > 2 then
    printf("hence the order of the LCLM is %d\n", ell);
  fi;
  if ell < s then
    for kk to k do
      S[kk] := MSylvester(LL[kk], ell, x, M, b);
    od;
    ID := MSylvester(-1, ell, x, M, b);
    Sigma := LinearAlgebra:-DiagonalMatrix([seq(S[kk], kk = 1..k)]);
    Sigma := <Sigma ; `<|>`(ID$k)>;

    rho := LinearAlgebra:-Rank(Sigma);
    if assigned('infolevel'[procname]) and 'infolevel'[procname] > 2 then
      printf("new matrix Sigma has size %d x %d and rank %d\n",
        LinearAlgebra:-Dimension(Sigma), rho)
    fi;
  fi;
  Sigma := LinearAlgebra:-Transpose(Sigma);
  K := LinearAlgebra:-NullSpace(Sigma);
  K := op(1, K);
  K := LinearAlgebra:-Transpose(K);
  m := (k + 1) * (ell + 1) - s;
  L := add(K[m - i]*M^i, i = 0..ell);
  L := normal(L);
  D := denom(L);      
  L := numer(L);
  c := content(L, M); 
  D := D/c;           
  L := collect(primpart(L, M), M, expand);
  if assigned('infolevel'[procname]) and 'infolevel'[procname] > 2 then
    printf("the degree of the LCLM is %d\n", degree(L, x))
  fi;
  if _params[5] <> 'undefined' then
    a[0] := 0;
    for j to k do
      a[j] := a[j - 1] + ell - rr[j] + 1;
      U[j] := add(normal(D*K[i])*M^(a[j] - i), i = a[j-1] + 1.. a[j])
    od;
    assign(_params[5], [seq(U[kk], kk = 1..k)])
  fi;
  return L
end proc:# LMOpLCLMBCLS12
#
LMOpLCLMBostanChyzakLiSalvy2012 := proc(Llist, x, M, b,
                                          cof::name:='undefined')
  option procname;
  local k, k1, L1, L2, cof1, cof2, res, c;
  k := nops(Llist);
  if k = 1 then
    if _params[5] <> 'undefined' then
      assign(_params[5], [1])
    fi;
  return Llist[1]
  fi;
  if k = 2 then
    return LMOpLCLMBCLS12(_params[1..-1])
  fi;
  k1 := iquo(k, 2);
  L1 := thisproc(Llist[1..k1], _params[2..-1]);
  if _params[5] <> 'undefined' then cof1 := eval(cof) fi;
  L2 := thisproc(Llist[k1 + 1..k], _params[2..-1]);
  if _params[5] <> 'undefined' then cof2 := eval(cof) fi;
  res := LMOpLCLMBCLS12([L1, L2], _params[2..-1]);
  if _params[5] <> 'undefined' then
    assign(cof,
      [seq(LMOpMul(cof[1], c, x, M, b), c = cof1),
       seq(LMOpMul(cof[2], c, x, M, b), c = cof2)])
  fi;
  return res
end proc:# LMOpLCLMBostanChyzakLiSalvy2012
#
MFlattening := proc(L, n, x, M)
  local LL, r, msg, j;
  option procname;
  LL := collect(L, M);
  r := degree(LL, M);
  if n < r then
    msg := "the second argument n = %1 is expected to be not smaller "
          "than the order of the first argument r = %2";
    error msg, n, r
  fi;
  return [seq(coeff(LL, M, n-j), j=0..n)]
end proc:# MFlattening
#
MSylvester := proc(L, n, x, M, b)
  local LL, r, i;
  option procname;
  LL := collect(L, M);
  r := degree(LL, M);
  Matrix(n-r+1, n+1,
    [seq(MFlattening(LMOpMul(M^(n-r-i), LL, x, M, b), n, x, M),
        i=0..n-r)])
end proc:# MSylvester
#
LMOpLCLMDumas1993 := proc(Llist, x, M, b, cof::name:='undefined')
  option procname;
  local k, k1, L1, L2, cof1, cof2, c, res;
  k := nops(Llist);
  if k = 1 then
    if _params[5] <> 'undefined' then
      assign(_params[5], [1])
    fi;
  return Llist[1]
  fi;
  if k = 2 then
    return LMOpLCLMD93(Llist[1], Llist[2], _params[2..-1])
  fi;
  k1 := iquo(k, 2);
  L1 := thisproc(Llist[1..k1], _params[2..-1]);
  if _params[5] <> 'undefined' then cof1 := eval(cof) fi;
  L2 := thisproc(Llist[k1 + 1..k], _params[2..-1]);
  if _params[5] <> 'undefined' then cof2 := eval(cof) fi;
  res := LMOpLCLMD93(L1, L2, _params[2..-1]);
  if _params[5] <> 'undefined' then
    assign(cof,
      [seq(LMOpMul(cof[1], c, x, M, b), c= cof1),
       seq(LMOpMul(cof[2], c, x, M, b), c= cof2)])
  fi;
  return res
end proc:# LMOpLCLMDumas1993
#
LMOpLCLMD93 := proc(L1, L2, x, M, b, cof::name:='undefined')
  local G, LL1, R, LL2, r1, r2, r, n, m, A, i, i1, z,
    K, msg, U1, U2, L, D, c;
  option procname;
  G := LMOpGCRD({L1, L2}, x, M, b);
  LL1, R := LMOpRightEuclideanDivision(L1, G, x, M, b);
  LL2, R := LMOpRightEuclideanDivision(L2, G, x, M, b);

  LL1 := numer(normal(LL1));
  LL1 := collect(LL1, M);
  if LL1 = 0 then return 0 fi;
  r1 := degree(LL1, M);
  LL2 := numer(normal(LL2));
  LL2 := collect(LL2, M);
  if LL2 = 0 then return 0 fi;
  r2 := degree(LL2, M);

  r := r1 + r2;

  n := r + 1;
  m := n + 1; 
  A := Matrix(m, m);
  for i from 0 to n do
    i1 := i + 1;
    A[i1, 1] := LMOpEval(LL1, x^i, x, M, b);
    for z from 1 to r2 do
      A[i1, z+1] := MOp(A[i1, z], x, b)
    od;
    A[i1, r2+2] := LMOpEval(LL2, x^i, x, M, b);
    for z from r2+2 to n do
      A[i1, z+1] := MOp(A[i1, z], x, b)
    od;
  od;

  if assigned('infolevel'[procname]) and 'infolevel'[procname] > 2 then
    printf("matrix A has size %d x %d\n",
      LinearAlgebra:-Dimension(A))
  fi;
  K := LinearAlgebra:-NullSpace(A);
  if nops(K) <> 1 then
    msg := "something is wrong in %1, ";
    msg := cat(msg, "since the kernel is not of dimension 1");
    error msg, procname
  fi;
  K := op(1, K);
  U1 := add(K[i]*M^(i-1), i=1..r2 + 1);
  U2 := - add(K[i]*M^(i-r2-2), i=r2+2..m); 
  L := LMOpMul(U1, LL1, x, M, b);
  L := LMOpMul(L, G, x, M, b);
  if has(L, RootOf) then
    Normalizer := evala@Normal
  fi;
  L := Normalizer(L);
  if has(L, I) then
    Normalizer := evalc@normal
  fi;
  L := Normalizer(L);
  D := denom(L);
  L := numer(L);
  c := content(L, M);
  D := D/c;
  L := collect(primpart(L, M), M, expand);

  if _params[6] <> 'undefined' then
    assign(_params[6], [collect(D*U1, M, normal), collect(D*U2, M, normal)])
  fi;
  return L
end proc:# LMOpLCLMD93
#
LMOpToPolynomialSolutions := proc(L, g, x, M, b,
    {output_format::Or(identical(undefined),identical(basis)):='undefined',
     degree_max::Or(nonnegint,identical(undefined)):='undefined',
      direction::{identical(lower), identical(upper)}:='upper',
      degree_min::Or(integer,identical(undefined)):='undefined',
      minimal_accuracy::{name,symbol,identical(undefined)}:='undefined',
      free::symbol := `tools/genglobal`('_C'),
      no_denominator::boolean:=false} ,$)
  local r, k, ell, deg, ldeg, LL, char_list, deg_max,
    s, ss, d, char, D, E, nu, nu1, bound_deg,
    deg_list, min_deg, max_deg, Z, y::nothing, Y, n, LY,
    sys, i, inc, j, sol, ind, sY, DD, s_dim,
    degY, matY, c, den, ans, koeff0,
    Eq, Iter, kappa, iter_bound, z::nothing, i0, m0, m1, m, q, t::nothing,
    result, kof
    ;
  
  uses LinearAlgebra;
  
  if g <> 0 then
    error "Non homogeneous equation not yet taken into account"
  fi;
  if Normalizer(coeff(L, M, 0)) = 0 then
    LL := LMOpNormalization(L, x, M, b);
    return LMOpToPolynomialSolutions(LL, g, x, M, b, _options)
  fi;
  if output_format = 'basis' then
    result := LMOpToPolynomialSolutions(L, g, x, M, b,
      subsop(2 = 'undefined', _options['output_format']),
      _options['degree_max'],
      _options['direction'],
      _options['degree_min'],
      _options['minimal_accuracy'],
      _options['free'],
      _options['no_denominator']
    );
    ind := select(has, indets(result, name), free);
    s_dim := nops(ind);
    if 'direction' = 'lower' then
      if no_denominator then
        result := [op(1, result),
                    sort([seq(eval(op(2, result),
                                  subsop([i,2] = 1, [seq(c = 0, c = ind)])),
                               i = 1..s_dim)],
                          `=`('key',sp->SeriesValuation(series(sp, x, 3), x)))
                  ];
      else
        result :=   sort([seq(eval(result,
                                  subsop([i,2] = 1, [seq(c = 0, c = ind)])),
                               i = 1..s_dim)],
                          `=`('key',sp->SeriesValuation(series(sp, x, 3), x)));
      fi
    else
      result := sort([seq(eval(result,
                              subsop([i,2] = 1, [seq(c = 0, c = ind)])),
                           i = 1..s_dim)],
                      `=`('key',sp->degree(sp, x)));
    fi;
    return result
  fi;
  if L = 0 then
    error "a nonzero operator is expected, but %1 is received", L
  fi;
  r := degree(L, M):
  if r = 0 then return 0 fi; 
  for k from 0 to r do
    ell[k] := coeff(L, M, k):
    deg[k] := degree(ell[k], x);
    ldeg[k] := ldegree(ell[k], x);
  od:

  if 'direction' = 'upper' then 
    if degree_max = 'undefined' then
      char_list :=
        LMOpConvexHull(L, x, M, b, 'variable' = z,
                    _options['direction']);
      deg_max := -1;
      s := nops(char_list);
      for ss to s do
        d := - char_list[ss]:-slope;
        char := char_list[ss]:-charpoly;
        if type(d, nonnegint) and Normalizer(eval(char, z = 1)) = 0 then
          deg_max := d; 
        fi;
      od;
      if deg_max = -1 then 
        return 0
      fi;
    else
      deg_max := degree_max
    fi;
    D := deg_max;
    E := max(seq(deg[k] + b^k*D, k=0..r));
  else 
    char_list :=
      LMOpConvexHull(L, x, M, b, 'variable' = z,
                    _options['direction']);
    s := nops(char_list);
    nu := - char_list[1]:-slope;
    nu1 := floor(nu);
    if minimal_accuracy <> 'undefined' then 
      assign(minimal_accuracy, nu)
    fi;
    if degree_max = 'undefined' then
      bound_deg := max(nu1, Order);
    else
      bound_deg := max(nu1, degree_max);
    fi;

    deg_list := select(type, [seq(- char_list[ss]:-slope, ss = 1..s)], integer);

    if nops(deg_list) > 0 then
      min_deg := deg_list[-1];
      max_deg := deg_list[1];
    else
      if no_denominator then
        return [1, 0]
      else
        return 0
      fi
    fi;

    if degree_min <> 'undefined' then
      min_deg := degree_min; 
      if max_deg < min_deg then
        if no_denominator then
          return [1, 0]
        else
          return 0
        fi
      fi;
    fi;
    if min_deg <> 0 then
      LL := LMOpChange(L,
        {x = t, y(x) = t^(+min_deg)*z(t)}, x, M, b);
      Z := LMOpToLaurentSeriesSolutions(
            LL,
            0, t, M, b,
            subsop(2 = bound_deg - min_deg, _options['degree_max']),
            subsop(2 = 0, _options['degree_min']),
            _options['free'], _options['output_format'],
            _options['no_denominator']
            );
      if no_denominator then
        if min_deg > 0 then
          return [op(1, Z),
            series((x*op(1, Z))^min_deg*eval(op(2, Z), t = x), x,
                                                       bound_deg + 1)]
        else
          return [op(1, Z),
            series((x*op(1, Z))^min_deg*eval(op(2, Z), t = x), x,
                                             bound_deg - min_deg + 1)]
        fi
      else
        if min_deg > 0 then
          return series(x^min_deg*eval(Z, t = x), x, bound_deg + 1)
        else
          return series(x^min_deg*eval(Z, t = x), x, bound_deg - min_deg + 1)
        fi
      fi
    fi;

    D := nu1;
    E := nu1 + ldeg[0];
  fi;
  y := table();
  Y := add(y[n]*x^n, n = 0..D);
  LY := series(LMOpEval(L, Y, x, M, b), x, E + 1);
  LY := eval(LY, O = 0);
  sys := {seq(op(2*i - 1, LY), i = 1..iquo(nops(LY), 2))};
  inc := {seq(y[j], j = 0..D)};
  sol := solve(sys,inc);
  Y := eval(Y, sol);
  ind := select(has, indets(Y, name), y);
  Y := collect(Y, ind);
  sY := series(Y + O(x^(D + 1)), x, D  +1);
  DD := iquo(nops(sY), 2) - 1;

  s_dim := nops(ind);
  degY := table();
  matY := Matrix(s_dim, DD);
  for j to DD do
    degY[j] := op(2*j, sY);
    c := collect(op(2*j - 1, sY), ind);
    for i to s_dim do
      if 'direction' = 'lower' then
        matY[i, j] := coeff(c, op(i, ind), 1)
      else
        matY[i, DD - j + 1] := coeff(c, op(i, ind), 1)
      fi
    od;
  od;
  matY := ReducedRowEchelonForm(matY);

  if type(matY, 'Matrix'(rational)) then
    for i to s_dim do
      den := ilcm(seq(denom(matY[i,j]), j = 1..DD));
      matY := RowOperation(matY, i, den, 'inplace' = true)
    od;
  fi;

  if 'direction' = 'lower' then
    ans := add(free[i]*add(matY[i,j]*x^degY[j], j = 1..DD), i = 1..s_dim);
  else
    ans := add(free[i]*add(matY[i, DD - j + 1]*x^degY[j],
                                            j = 1..DD), i = 1..s_dim)
  fi;

  if 'direction' = 'upper' then
    return ans
  fi;

  if ans = 0 then
    if no_denominator then
      return [1, 0]
    else
      return 0
    fi
  fi; 
  koeff0 := coeff(ell[0], x, ldeg[0]);
  if has(koeff0, RootOf) then
    koeff0 := evala(koeff0)
  fi;
  if member(koeff0, {-1, 1}) then
      Eq := add(ell[k] * eval(ans + x^(D + 1)*z(x), x = x^(b^k)), k = 0..r);
      Eq := collect(Eq, z, Normalizer);
      Iter := solve(Eq, z(x));
      Iter := collect(Iter, z, Normalizer);
      Z := 0;
      for kappa while collect(coeff(L, M, kappa), x) = 0 do od;
      iter_bound := 1 +
        log[b](1 + (b - 1)/(b^kappa - 1)
                    * (bound_deg - nu1)/ (nu1 + 1 - nu));
      iter_bound := ceil(iter_bound);
      for j to iter_bound do
        Z := eval(Iter, {seq(z(x^(b^k)) = eval(Z, x = x^(b^k)), k = 1..r)});
        Z := series(Z, x, bound_deg + 1); 
        Z := convert(Z, polynom);
      od;
      result := series(ans + x^(nu1 + 1)*Z + O(x^(bound_deg + 1)),
                    x, bound_deg + 1);
      if no_denominator then
        return [1, result]
      else
        return result
      fi
  else
      kof := table();
      kof[0] := 1;
      for i to bound_deg do
        kof[i] := evala(koeff0 * kof[i - 1])
      od;

      Z := table();
      for n from 0 to nu1 do
        Z[n] := evala(coeff(ans, x, n)*kof[n])
      od;
      for i from nu1 + 1 to bound_deg do
        z := add(coeff(ell[0], x, j) * Z[ldeg[0] + i - j]
                    * kof[- ldeg[0] + j - 1], 
                  j = ldeg[0] + 1..min(ldeg[0] + i, deg[0]));
        for k from 1 to r do
          i0 := irem(ldeg[0] + i, b^k, 'q');
          m0 := ceil((ldeg[k] - i0)/b^k);
          m1 := min(q, floor((deg[k] - i0)/b^k));
          z := z + add(coeff(ell[k], x, i0 + m*b^k) * Z[q - m]
                          * kof[i - q + m - 1], m = m0..m1);
        od;
          Z[i] := evala(-z);
       od;
    if no_denominator then
      result :=  [koeff0,
              series(add(Z[i]*x^i        , i = 0..bound_deg)
                                      + O(1)*x^(bound_deg + 1),
                              x, bound_deg + 1)
              ];
      return result
    else
      return series(add(Z[i]*x^i/kof[i], i = 0..bound_deg)
                        + O(1)*x^(bound_deg + 1),
                x, bound_deg + 1)
    fi;
  fi;
end proc:# LMOpToPolynomialSolutions
#
LMOpToRationalFunctionsSolutions := proc(L, g, x, M, b,
    {output_format::Or(identical(undefined),identical(basis)):='undefined',
    output::{identical(ratpoly), identical(ramratpoly)}:='ratpoly',
    method::{identical(LMOpRatSol2018),identical(LMOpRatSol2022)}:=
      'LMOpRatSol2022',
    free::symbol := `tools/genglobal`('_C')
    } ,$)
  local LL, w, chi, q, LLL, res_basis, result, i, r, lambda::nothing;
  option procname;
  uses LinearAlgebra;
  LL := collect(L, M, Normalizer);
  w := ldegree(LL, M);
  if output = 'ramratpoly' then
    LL := normal(LL/M^w);
    chi := LMOpConvexHull(LL, x, M, b,
            'direction' = 'lower', 'variable' = lambda);
    q := WellChosenDenominator([seq(r:-slope, r = chi)], b);
    LLL := LMOpChange(LL, {x = x^q, y(x) = z(x)}, x, M, b);
    res_basis := method(LLL, g, x, M, b);
    res_basis := map(simplify,
                        eval(res_basis, x = x^(1/(q*b^w))),
                        'power', 'symbolic');
    if output_format = 'basis' then
      return res_basis
    else
      result := add(free[i]*res_basis[i], i = 1..nops(res_basis));
      return result
    fi;
  fi;
  if w > 0 then
    LL := LMOpNormalization(LL, x, M, b)
  fi;
  res_basis := method(LL, g, x, M, b);
  if output_format = 'basis' then
    result := res_basis
  else
    result := add(free[i]*res_basis[i], i = 1..nops(res_basis));
  fi;
  return result
end proc:# LMOpToRationalFunctionsSolutions
#
LMOpRatSol2018 := proc(L, g, x, M, b)
  option procname;
  local LL, r, k, ell, delta, br1, v_bar, q_star, MOpkq_star, e,
    elltilde, i, Ltilde, polbasis, sigma, j;

  if g <> 0 then
    error "Non homogeneous equation not yet taken into account"
  fi;
  LL := collect(L, M);
  if ldegree(LL, M) > 0 then
    LL := LMOpNormalization(L, x, M, b)
  fi;
  r := degree(LL, M);
  for k from 0 to r do
    ell[k] := coeff(LL, M, k)
  od;
  delta := max(seq(degree(ell[k], x), k = 0..r));
  br1 := b^(r-1);
  if delta < br1 then
    if expand(add(ell[k], k = 0..r)) = 0 then
      return [1]
    else
      return []
    fi
  fi;
  v_bar := floor(delta/br1/(b - 1));
  q_star := ParamountDenominator(x, b, r, ell[r]);
  for k from 0 to r do
    MOpkq_star[k] := MOp(q_star, x, b^k)
  od;
  for k from 0 to r do
    e[k] := floor(b*delta/(b - 1)) - b^k*v_bar;
    elltilde[k] := x^e[k]*ell[k]*
      mul(MOpkq_star[i], i = 0..(k-1))*
      mul(MOpkq_star[i], i = (k+1)..r)
  od;
  Ltilde := add(elltilde[k]*M^k, k = 0..r);
  polbasis := LMOpToPolynomialSolutions(Ltilde, 0, x, M, b,
    'degree_max' = degree(q_star) + 2*v_bar + 1,
    'output_format' = 'basis');
  sigma := nops(polbasis);
  return [seq(normal(polbasis[j]/x^v_bar/q_star), j = 1..sigma)]
end proc:# LMOpRatSol2018
#
ParamountDenominator := proc(x, b, r, ellr)
  option procname;
  local ell, btother, result, k, u, s, uu, v, i, btotherminus1,
    utilde, Gutilde, v_lemma_2_2;
  if r = 0 then return 1 fi;

  ell := ellr;
  btother := b^r;
  result := 1;
  for k do
    u := 0;
    for s from 0 to btother - 1 do
      u := gcd(u, SOp(ell, x, btother, s));
    od;
    if degree(u, x) = 0 then break fi;
    result := result * u;
    uu := u;
    v := uu;
    for i to r - 1 do
      uu := MOp(uu, x, b);
      v := lcm(v, uu)
    od;
    uu := MOp(uu, x, b);
    ell := quo(ell, uu, x)*v
  od;
  btotherminus1 := b^(r-1);
  utilde := 0;
  for s from 0 to btotherminus1 - 1 do
    utilde := gcd(utilde, SOp(ell, x, btotherminus1, s));
  od;
  Gutilde := GOp(utilde, x, b);
  result := result * Gutilde;

  v_lemma_2_2 := floor(ldegree(ellr, x)/b^(r-1)/(b-1));

  return x^v_lemma_2_2*result
end proc:# ParamountDenominator
#
LMOpRatSol2022 := proc(L, g, x, M, b,
    {output_format::Or(identical(undefined),identical(basis)):='basis',
    sigma_min::Or(posint,identical(undefined)):='undefined',
    sigma_ratio::Or(realcons,identical(golden_ratio),
                     identical(radix)):='golden_ratio',
    sigma_shift::Or(posint,identical(undefined)):='undefined',
    sigma_max::Or(posint,identical(infinity)):='infinity',
    free::symbol := `tools/genglobal`('_C')
    } ,$)
  local phi := 1.61803398874989484820,
    LL, r, k, ell, deg, ldeg, result, sigma_multiplier, sigma_adder,
    F, t_dim, tt_dim, alpha, ancillary, deg_max, ancillary1, deg_ell_r,
    bound_B_den, bound_B_num, sigma, history, sigma_vect, i,
    max_degree_vect, f, sol, tried_sigma, next_sigma, den, F1, W, deg_W,
    rho, j, Delta, jj, candidate, free_seq, WP, choice, K, ind, s_dim, c
    , nu::nothing, t::nothing, tau::nothing;
  option procname;
  uses LinearAlgebra;
  LL := collect(L, M, Normalizer);

  Order := 10;
  r := degree(LL, M):
  for k from 0 to r do
    ell[k] := coeff(LL, M, k):
    deg[k] := degree(ell[k], x);
    ldeg[k] := ldegree(ell[k], x);
  od:
  if r < 1 then
    result := [];
    return result;
  fi;

  if sigma_shift = 'undefined' then
    if type(sigma_ratio, realcons) then
     if member(sigma_ratio, {+infinity, -infinity}) then
       error "the option sigma_ratio must provide a real number"
     fi;
     sigma_multiplier := evalf(sigma_ratio, 10);
     if sigma_multiplier <= 1. then
       error "the option sigma_ratio must provide a real number > 1"
     fi;
    elif sigma_ratio = 'golden_ratio' then
     sigma_multiplier := phi
    elif sigma_ratio = 'radix' then
     sigma_multiplier := b
    else
     error "option sigma_ratio not well defined"
    fi;
  else
    if sigma_max = 'infinity' then
     error "sigma_max should have an integer value if sigma_shift is used"
    fi;
    sigma_adder := sigma_shift
  fi;

  F := LMOpToLaurentSeriesSolutions(LL, 0, x, M, b,
           subsop(2 = 'basis', _options['output_format']));
  t_dim := nops(F);
  tt_dim := t_dim + 1;

  if t_dim = 0 then
    result := [];
    return result;
  fi;
  alpha := SeriesValuation(F[1], x);
  ancillary := LMOpChange(L, {x = x, y(x) = x^(+alpha)*z(x)}, x, M, b);
  deg_max := max(seq(degree(coeff(ancillary, M, k), x), k = 0..r));
  if deg_max < b^r - 1 then
    ancillary1 := expand(add(coeff(ancillary, M, k), k = 0..r));
    if ancillary1 = 0 then
    result := [x^alpha];
    return result;
    fi;
  fi;
  F := LMOpToLaurentSeriesSolutions(ancillary, 0, x, M, b,
        _options['free'], 'minimal_accuracy' = 'nu');
  deg_ell_r := degree(coeff(ancillary, M, r), x);
  bound_B_den := floor(min(
    3*deg_ell_r/b^r,
    deg_max*(b - 1)/(b^(r+1) - 2*b^r + 1)
  ));
  bound_B_num := bound_B_den + floor(deg_max/b^(r-1)/(b - 1));
  result := 0;
  if sigma_min <> 'undefined' then
    sigma := max(floor(nu) + 1, sigma_min)
  else
    sigma := max(floor(nu) + 1, 100)
  fi;
  history := table();
  for k do
    if sigma_shift = 'undefined' then
      if k > 3 and history[k - 1][2] > bound_B_num/5 then
        sigma_vect := Vector([seq(history[i][1], i = k - 3..k - 1)]);
        max_degree_vect := Vector([seq(history[i][2], i = k - 3..k - 1)]);
        f := Statistics:-Fit('intercept' + 'slope'*tau,
              sigma_vect, max_degree_vect, tau);
        sol := solve(f = bound_B_num + 10, t);
        if type(sol, rational) then
          tried_sigma := ceil(sol);
        else
          tried_sigma := 0
        fi;
        if type(tried_sigma, posint) and tried_sigma > sigma then
          sigma := tried_sigma
        fi;
      fi;
      next_sigma := floor(sigma * sigma_multiplier)
    else
      next_sigma := sigma + sigma_adder
    fi;
    if sigma > sigma_max then
      return FAIL
    fi;

    F := LMOpToLaurentSeriesSolutions(
            ancillary, 0, x, M, b,
            _options['free'], 'degree_min' = 0,
            subsop(2 = 'basis', _options['output_format']),
            'degree_max' = sigma,
            'no_denominator' = true);

    den := F[1];
    F1 := [op(F[2]), 1];
    W := HermitePadeApproximants(F1, x, sigma + 1);
    if W = FAIL then return FAIL fi;

    deg_W := map(degree, W, x);

    history[k] := [sigma, max(op(convert(deg_W, list)))];
    W := SubMatrix(W,
      select( proc(i) local j;
                evalb(max(seq(deg_W[i, j], j = 1..(tt_dim)))
                        <= bound_B_num)
              end,[$1..(tt_dim)]), [$1..(tt_dim)]);
    rho := RowDimension(W);

    if rho = 0 then
      break
    fi;
    if rho = tt_dim then
      sigma := next_sigma;
      next
    fi;
    if rho = t_dim then
      for j to tt_dim  do
        Delta[j] :=
          Determinant(SubMatrix(W, 1..rho,
            [seq(jj, jj = 1..j-1), seq(jj, jj = j+1..tt_dim)]));
      od;
      candidate := normal(
          add(free[j]*(-1)^(j)*Delta[j], j = 1..t_dim)/
          (-1)^tt_dim/Delta[tt_dim]
        );
    else
      free_seq := seq(free[i], i = 1..t_dim);
      WP := < W, `<|>`(free_seq, 0),
                 `<|>`(0$t_dim, 1)>;
      choice := combinat:-choose(tt_dim, rho + 2);
      for i to binomial(tt_dim, rho + 2) do
        Delta[i] :=
          Determinant(SubMatrix(WP, 1..rho+2, choice[i]));
        Delta[i] := {coeffs(collect(Delta[i], x), x)};
      end do;
      sol := solve(`union`(seq(Delta[i], i = 1..binomial(tt_dim, rho + 2))),
                   {free_seq} );
      if eval({free_seq}, sol) = {0} then
        next k
      fi;
      K := NullSpace(eval(Transpose( WP), sol));
      if nops(K) > 1 then
        sigma := next_sigma;
        next k
      fi;
      K := op(K);
      if K[rho + 1] = 0 or K[rho + 2] = 0 then
        sigma := next_sigma;
        next k
      fi;
      candidate :=normal(- K[rho + 2]/K[rho + 1]);
    fi;;
    if degree(numer(candidate), x) > bound_B_num
      or degree(denom(candidate), x) > bound_B_den then
      sigma := next_sigma;
      next k
    fi;
    candidate := evala(eval(candidate, x = x/den));
    if LMOpEval(ancillary, candidate, x, M, b) = 0 then
      result := candidate
    else
      sigma := next_sigma;
      next k
    fi;
    break
  od;

  ind := select(has, indets(result, name), free);
  s_dim := nops(ind);
  result := sort([seq(eval(result, subsop([i,2] = 1, [seq(c = 0, c = ind)])),
                  i = 1..s_dim)],
                  `=`('key',sp->SeriesValuation(series(sp, x, 3), x)));
  result := map(foo -> normal(x^alpha*foo), result);
  return result
end proc:# LMOpRatSol2022
#
LMOpToLaurentSeriesSolutions := proc(L, g, x, M, b,
    {output_format::Or(identical(undefined),identical(basis)):='undefined',
      degree_max::nonnegint:=Order,
      degree_min::Or(integer,identical(undefined)):='undefined',
      minimal_accuracy::{name,symbol,identical(undefined)}:='undefined',
      free::symbol := `tools/genglobal`('_C'),
      no_denominator::boolean:=false} ,$)
  option procname;
  LMOpToPolynomialSolutions(L, g, x, M, b,
          _options['output_format'],
          _options['degree_min'], _options['degree_max'],
          _options['minimal_accuracy'],
          _options['free'], 'direction' = 'lower',
          _options['no_denominator']);
end proc:# LMOpToLaurentSeriesSolutions
#
LMOpToPuiseuxSeriesSolutions := proc(L, g, x, M, b,
  {output_format::Or(identical(undefined),identical(basis)):='undefined',
  degree_max::nonnegint:=Order,
  field_extension::Or(identical(splitting_field),
                      radalgnum,set(radalgnum)):={},
  lambda_choice::Or(set,identical(undefined)):='undefined',
  free::symbol := `tools/genglobal`('_C'),
  ancillary_eqs::boolean:=false},
  $)
  local LL, w, res_basis, res, lambda, q, j, chi, r,
    Zeta, lower_admissible_edges, Lambda_star, LLL, Y,
    y::nothing, z::nothing, ancillary, tosubs, invtosubs, nu::nothing, omega;
  option procname;

  w := ldegree(L, M);
  LL := normal(L/M^w);

  if output_format = 'undefined' then
    res_basis := LMOpToPuiseuxSeriesSolutions(LL, g, x, M, b,
      subsop(2 = 'basis', _options['output_format']),
      subsop(2 = b^w*degree_max, _options['degree_max']),
      _options['field_extension'],
      _options['lambda_choice'],
      _options['ancillary_eqs'],
      _options['degree_max']);
    res := 0;
    
    
    for lambda in res_basis['roots'] do
      q := res_basis[lambda][1];
      omega := min(op(map2(op, -1, res_basis[lambda][2])));
      res := res + ln(1/x)^(log[b](lambda)) *
        PuiseuxSeries(q*b^w,
          series(add(free[lambda, j] * res_basis[lambda][2][j],
                      j = 1..nops(res_basis[lambda][2])),
                 x, omega))
    od;
    return res
  fi;

  if field_extension = 'splitting_field' then
    chi := LMOpConvexHull(L, x, M, b,
        'direction' = 'lower', 'variables' = lambda);
    Zeta := FieldExtension2({seq(r:-charpoly, r = chi)}, lambda,
                'tosubs', 'invtosubs')
  else
    Zeta := FieldExtension1(field_extension, 'tosubs', 'invtosubs')
  fi;
  invtosubs := invtosubs
    union {RootOf(_Z^2 + 1) = I, RootOf(_Z^2 + 1, 'index' = 1) = I};
  if Zeta <> {} and has(Zeta, RootOf) then
    Normalizer := (evala@Normal)
  fi;

  lower_admissible_edges :=
    AdmissibleEdges(L, x, M, b, _options['field_extension']);
  Lambda_star := map(op,{indices(lower_admissible_edges)});
  if lambda_choice <> 'undefined' then
    Lambda_star := Lambda_star intersect lambda_choice
  fi;
  q := table();
  res := table();
  res['roots'] := Lambda_star;
  for lambda in Lambda_star do
    q[lambda] :=
      WellChosenDenominator(map2(op, 2, lower_admissible_edges[lambda]), b);
    LLL := evala(eval(
            LMOpChange(LL, {x = x^q[lambda], y(x) = z(x)}, x, M, b),
            M = lambda*M));
      Y := LMOpToLaurentSeriesSolutions(LLL, 0, x, M, b,
             subsop(2 = q[lambda]*degree_max, _options['degree_max']),
             _options['output_format'], 'minimal_accuracy' = 'nu');
      if has(Y, RootOf) then
        Y := evala(Y)
      fi;
      if ancillary_eqs then
        ancillary[lambda] := LLL
      else
        ancillary[lambda] := NULL
      fi;
        res[lambda] :=
          [q[lambda], Y, nu, ancillary[lambda]]
  od;
  return
    table(eval([seq(lambda = res[lambda],
                  lambda = Lambda_star),
                'roots' = Lambda_star], invtosubs))

end proc:# LMOpToPuiseuxSeriesSolutions
#
# change of variables in a homogeneous linear Mahler equation
LMOpChange := proc(L::polynom(anything,{x,M}),
          c::{equation, list(equation), set(equation)},
          x::name, M::name, b::radix, $)
  local s, sv, mon, sf, LL, gamma, yofx, zoft, t, beta, alpha, r,
    k, ellprime, g, msg;
  if type(c,':-set') then 
    s := convert(c,'list')
  elif type(c,'equation') then
    s := [c]
  else
    s := c
  end if;
  sv := select(proc(eq) evalb(op(1,eq) = x) end,s); 
  if nops(sv) <> 1 then
    error "exactly one equation of the form %1 = ... is expected", x
  fi;
  mon := op([1,2],sv); 
  sf := select( proc(eq)
                  local eqrhs;
                  eqrhs := op(1,eq);
                  evalb(type(eqrhs,function) and [op(eqrhs)] = [x])
                end,s);
  if nops(sf) = 0 then
    msg := cat(
      "expecting equations of the form %1 = ... and f(%1) = ... ",
      "but received %2"
    );
    error msg, x, c
  fi;
  if nops(sf) <> 1 then
    msg := cat(
      "expecting exactly one equation of the form f(%1) = ... ",
      "but received %2"
    );
    error msg, x, sf
  fi;
  yofx := op([1,1],sf);
  if not (type(yofx,function) and [op(1,yofx)] = [x]) then
    msg := cat(
      "incoherent change of variable, the variable in %1 must be %2"
    );
    error msg, yofx, x
  fi;
  zoft := op([1,2],sf);
  zoft := op(indets(zoft,function));
  t := [op(zoft)];
  if nops(t) <> 1 then
    msg := cat(
      "expecting one function of one argument, but received %1"
    );
    error msg, zoft
  fi;
  t := op(t); 

  if (not (type(mon,`^`) and type(op(2,mon),posint)))
    and (not mon = 1) and (not mon = t) then
    msg := cat(
      "a monomial with positive integer exponent is expected ",
      "in the equation %1 = ..., but received %2"
    );
    error msg, x, x = mon
  fi;
  if mon = 1 then
    beta := 0
  elif mon = t then
    beta := 1
  else
    beta := op(2,mon); 
  fi;
  mon := op([1,2],sf);
  mon := normal(mon/zoft);
  if (not (type(mon,`^`) and type(op(2,mon),integer)))
    and (not mon = 1) and (not mon = t) then
    msg := cat(
      "a monomial with integer exponent is expected in ",
      "the equation %1 = ... (as a cofactor of %3), but received %2"
    );
    error msg,yofx, yofx = mon * zoft, zoft
  fi;
  if mon = 1 then
    alpha := 0
  elif mon = t then
    alpha := 1
  else
    alpha := op(2,mon); 
  fi;

  LL := collect(L,M);
  r := degree(LL,M);
  for k from 0 to r do
    ellprime[k] := normal(t^(alpha*b^k)*eval(coeff(LL,M,k),x=t^beta));
    g[k] := ldegree(ellprime[k],t)
  od;
  gamma := min(seq(g[k],k=0..r));
  userinfo(3,'procname',"gamma = ", gamma);

  return add(normal(t^(-gamma)*ellprime[k])*M^k,k=0..r)
end proc:# LMOpChange
#
# picture of the Newton polygon of a linear Mahler operator
LMOpNewtonPolygon := proc(
    L::polynom(anything,{x,M}), g, x::name, M::name, b::radix)
  local LL, P, Up, Lo, pixP, pixL, pixU, opt,
    abscissamax, ordinatemax, m, pt, abscissamin, ordinatemin, msg;

  LL := collect(L, {x,M}, 'distributed');
  P := LMOpSupport(LL, x, M, b); 

  if g <> 0 then
    msg := "not yet implemented in the nonhomogeneous case";
    error msg
  fi;

  abscissamax := max(seq(pt[1],pt=P));
  ordinatemax := max(seq(pt[2],pt=P));
  abscissamin := 0;
  ordinatemin := min(0, seq(pt[2],pt=P));
  userinfo(3, 'procname', "Newton diagram of the Mahler operator: ",
    convert(P,string));
  Lo, Up := LMOpConvexHull(LL, x, M, b, 'output' = 'geometry');
  userinfo(3,'procname',"lower convex hull: ",convert(Lo,string));
  userinfo(3,'procname',"upper convex hull: ",convert(Up,string));
  opt := _passed[6..-1];
  pixP := plot(P, 'color'='black','style'='point','symbol'='solidcircle', opt):
  pixL := plot(Lo, 'color'='blue', opt):
  pixU := plot(Up, 'color'='red', opt):
  plots:-display({pixP,pixL,pixU},
    'view'=[abscissamin..abscissamax,ordinatemin..ordinatemax],opt)

end proc:# LMOpNewtonPolygon
#
LMOpSupport := proc(L, x, M, b)
  local Mon, z, alpha;
  Mon := LMOpMonomials(_passed);
  `union`(seq(map(z->[b^alpha,z], Mon[alpha]), 
            alpha = map(op,[indices(Mon)])));
end proc:# LMOpSupport
cachetableLMOpSupport := Cache(100):
Cache(cachetableLMOpSupport, procedure=LMOpSupport):
#
LMOpMonomials := proc(L, x, M, b)
  local LL, rplus, rmoins, alpha, K, koeff, m;
  LL := collect(L, {x,M}, 'distributed');
  rplus := degree(LL, M);
  rmoins := ldegree(LL, M);
  for alpha from rmoins to rplus do
    koeff := coeff(L, M, alpha);
    if koeff <> 0 then
      if type(koeff, 'monomial') then
        K[alpha] := {degree(koeff,x)}
      else
        K[alpha] := {seq(degree(m, x), m = koeff)}
      fi
    fi
  od;
  return eval(K)
end proc:# LMOpMonomials
cachetableLMOpMonomials := Cache(100):
Cache(cachetableLMOpMonomials, procedure=LMOpMonomials):
#
# solving of linear homogeneous Mahlerian equations
LMOpSolve := proc(L::polynom(ratpoly(anything, x), M),
                    x::{name, symbol}, M::{name, symbol}, b::radix,
     {output_format::Or(identical(undefined),identical(basis)):='undefined',
      output::Or(identical(polynom),
                identical(ratpoly),
                identical(ramratpoly),
                identical(mhypergeom),
                identical(series),
                identical(laurentseries),
                identical(puiseuxseries),
                identical(extendedpuiseuxseries),
                identical(riccati))
                :='extendedpuiseuxseries',
      free::symbol := `tools/genglobal`('_C'),
      field_extension::Or(identical(splitting_field),
                          radalgnum,set(radalgnum)):={},
      righthand_side::ratpoly:=0})
  local LL, result;
  if normal(righthand_side) <> 0 then
    error "dealing with inhomogeneous equations not yet implemented"
  fi;
  if Normalizer(L) = 0 then
    error "a nonzero operator is expected, but %1 is received", L
  fi;
  LL := L;
  LL := numer(normal(LL));
  LL := collect(LL, {x, M}, 'distributed');
  if 'output' = 'polynom' then
    result :=  LMOpToPolynomialSolutions(LL, righthand_side, x, M, b,
            'direction' = 'upper',
            _options['free'],
            _options['output_format'])
  elif 'output' = 'ratpoly' or 'output' = 'ramratpoly' then
    result :=  LMOpToRationalFunctionsSolutions(LL, righthand_side, x, M, b,
            _options['free'],
            _options['output_format'],
            _options['output']) 
  elif 'output' = 'mhypergeom' then
    result :=  LMOpToMHypergeometricSolutions(LL, x, M, b,
                  _options['output_format'],
                  _options['free'],
                  _options['field_extension'])
  elif 'output' = 'series' then
    result :=  LMOpToLaurentSeriesSolutions(LL, righthand_side, x, M, b,
            'degree_min' = 0, 'degree_max' = Order,
            _options['free'],
            _options['output_format'])
  elif 'output' = 'laurentseries' then
    result :=  LMOpToLaurentSeriesSolutions(LL, righthand_side, x, M, b,
              'degree_max' = Order,
              _options['free'],
              _options['output_format'])
  elif 'output' = 'puiseuxseries' then
    result :=  LMOpToPuiseuxSeriesSolutions(LL, righthand_side, x, M, b,
            'lambda_choice' = {1},
            'degree_max' = Order,
            _options['free'],
            _options['output_format'])
  elif 'output' = 'extendedpuiseuxseries' then
    result :=  LMOpToPuiseuxSeriesSolutions(LL, righthand_side, x, M, b,
            'degree_max' = Order,
            _options['field_extension'],
            _options['free'],
            _options['output_format'])
  elif 'output' = 'riccati' then
    result := eval(RiccatiSolve(LL, x, M, b,
            _options['field_extension'],
               _options['free']))
  else
    error "unknown option: output = ", output
  fi;
  return eval(result) 
end proc:# LMOpSolve
#
# truncation of a linear Mahler operator
LMOpTruncation := proc(L::polynom(polynom(anything, x), M),
  x::name, M::name, b::radix, {slope::rational:=0},$)
  local LL, r, k, c, v, nu, mu, j;
  LL := collect(L, M, expand);
  r := degree(LL, M);
  for k from 0 to r do
    c[k] := coeff(L, M, k);
    v[k] := ldegree(c[k], x);
  od;
  nu := max(seq((v[0] - v[k])/(b^k - 1), k = 1..r));
  mu := nu + v[0];
  add(add(coeff(c[k], x, j)*x^j,
    j = v[k]..floor(mu + slope*b^k))*M^k, k = 0..r)
end proc:# LMOpTruncation
#
# first order right factors of a linear Mahler operator
RiccatiSolve := proc(L::polynom(ratpoly(anything, x), M),
  x::name, M::name, b::radix,
    {
    field_extension::Or(identical(splitting_field),
                        radalgnum,set(radalgnum)):={},
    sigma_min::Or(posint,identical(undefined)):='undefined',
    sigma_ratio::Or(realcons,identical(golden_ratio)):='golden_ratio',
    sigma_shift::Or(posint,identical(undefined)):='undefined',
    sigma_max::Or(posint,identical(infinity)):='infinity',
    free::symbol := `tools/genglobal`('_C'),
    method::Or(identical(basic_petkovsek),identical(improved_petkovsek),
      identical(hermite_pade_approximants)):='hermite_pade_approximants',
    prime_decomposition::Or(identical(Maple),identical(Singular)):='Singular'
    })
  local LL, w, r, res, __C
    ;
  LL := numer(normal(L));
  LL := collect(LL, {x, M}, 'distributed');
  r := degree(LL, M);
  w := ldegree(LL, M);
  LL := normal(LL / M^w);
  if r < 1 then res := {}
  elif r = 1 then res := {Normalizer(- coeff(LL, M, 0)/coeff(LL, M, 1))}
  else
    if method = 'basic_petkovsek' or method = 'improved_petkovsek' then
      res := RiccatiByPetkovsek(LL, x, M, b,
                _options['free'],
                _options['field_extension'], _options['method'])
    elif method = 'hermite_pade_approximants' then
      res := RiccatiByHermitePade(LL, x, M, b,
                _options['free'],
                _options['field_extension'],
                _options['sigma_ratio'], _options['sigma_shift'],
                _options['sigma_min'],   _options['sigma_max'],
                _options['prime_decomposition'])
    else error "unknown method: %1", method fi
  fi;
  __C := `tools/genglobal`('_C');
  res := eval(res, {x = x^(1/b^w), __C = free});
  if type(w, 'posint') then res := res union {0} fi;
  return res
end proc:# RiccatiSolve
#
RiccatiByPetkovsek := proc(L, x, M, b, {
  field_extension::Or(identical(splitting_field),
                      radalgnum,set(radalgnum)):={},
  method::Or(identical(basic_petkovsek),identical(improved_petkovsek)):=
    'improved_petkovsek',
  free::symbol := `tools/genglobal`('_C')})
  local LL, chi, alg_ext, lower_admissible_edges, Lambda, q_table,
    mu, q_L, tosubs, invtosubs, lambda, lower_edges_slopes, r
    ;
  LL := L;

  alg_ext := FieldExtension1(field_extension, 'tosubs', 'invtosubs');
  alg_ext := eval(alg_ext, tosubs);
  if has(alg_ext, RootOf) then
    Normalizer := (evala@Normal)
  fi;

  chi := LMOpConvexHull(LL, x, M, b,
    'direction' = 'lower', 'variable' = 'lambda');
  if 'field_extension' = 'splitting_field' then
    lower_edges_slopes := [seq(r:-slope, r = chi)];
    q_L := WellChosenDenominator(lower_edges_slopes, b)
  else
    lower_admissible_edges :=
      AdmissibleEdges(LL, x, M, b,
        subsop(2 = alg_ext, _options['field_extension']));
    Lambda := map(op,{indices(lower_admissible_edges)});
    q_table := table();
    for mu in Lambda do
      q_table[mu] :=
        WellChosenDenominator(map2(op, 2, lower_admissible_edges[mu]), b);
    od;
    q_L := ilcm(seq(q_table[mu], mu=Lambda));
  fi;
  if method = 'basic_petkovsek' then
    return eval(
            BasicPetkovsek(LL, x, M, b, q_L,
             _options['field_extension'], _options['free']),
            invtosubs)
  else
    return eval(
            ImprovedPetkovsek(LL, x, M, b, q_L,
             _options['field_extension'], _options['free']),
            invtosubs)
  fi
end proc:# RiccatiByPetkovsek
#
ImprovedPetkovsek := proc(L, x, M, b, q_L, {
  field_extension::Or(identical(splitting_field),
                      radalgnum,set(radalgnum)):={},
  free::symbol := `tools/genglobal`('_C')})
  local  LL, r, k, ell, res, candidate_table, candidate_counter,
    br1, zeta, upper_admissible_edges, Zeta_star,
    uae, MSF, degree_table, i, q, p_divides_Mp, msfMp, p,
    GROp_lcm, known, old, j, tb, Gq, jj, GROp_r_minus_1, qq, Q_i_set,
    Q_o_set, F_table, a_check, b_check, b_min, B_start, B_finish,
    B, deg_B, bool, Q, a_bar, a_max, A_start, A_finish,
    A, deg_A, polyA, polyB, Ltilde, nuae, deg_max, polyC,
    __C::nothing, alg_ext, t::nothing, tosubs, invtosubs
  ;
  LL := L;
  alg_ext := indets([LL, _options['field_extension']], {'radical', 'RootOf'});
  FieldExtension1(alg_ext, 'tosubs', 'invtosubs');
  LL := eval(LL, tosubs);
  if has(LL, RootOf) then
    Normalizer := (evala@Normal);
    LL := collect(LL, M, Normalizer);
  fi;


  upper_admissible_edges := AdmissibleEdges(L, x, M, b,
      'direction' = 'upper', _options['field_extension']);
  Zeta_star := map(op,{indices(upper_admissible_edges)});

  if Zeta_star = {} then
    return {}
  else
  fi;
  LL := primpart(L, M);
  LL := eval(LL, x = t^(q_L));

  r := degree(LL, M);
  for k from 0 to r do
    ell[k] := coeff(LL, M, k)
  od;
  candidate_table := table();
  candidate_counter := 0;
  br1 := b^(r - 1);



  MSF := table();
  degree_table := table();
  for i in [0,r] do
    MSF[i] := RatPolyToMSF(ell[i], t,
      _options['field_extension'],
      'direction' = 'upper');
    for q in MSF[i]['factors'] do
      degree_table[q] := degree(q, t)
    od
  od;

  p_divides_Mp := table();
  msfMp := table();
  for i in [0, r] do
    for p in MSF[i]['factors']  do
      if p = t then
        p_divides_Mp[p] := [false];
        next
      fi;

      msfMp[p] := RatPolyToMSF(MOp(p, t, b), t,
             _options['field_extension'],
             'direction' = 'upper');

      q := normal(MOp(p, t, b) / p);
      if type(q, 'polynom'('anything', t)) then
         p_divides_Mp[p] := [true,
           RatPolyToMSF(q, t,
             _options['field_extension'],
             'direction' = 'upper')]
      else
        p_divides_Mp[p] := [false]
      fi
    od
  od;

  GROp_lcm := table();
  for p in MSF[0]['factors'] do
    known := table();
    old := {p};
    known[p] := true;
    to r - 1 do
      j := 0;
      tb := table();
      for q in old do
        Gq := GOp(q, t, b);
        Gq := factors(Gq);
        Gq := op([2, 1, 1], Gq); 
        if assigned(known[Gq]) then
        else
          j := j + 1;
          tb[j] := Gq;
          known[Gq] := true
        fi;
      od;
      old := {seq(tb[jj], jj = 1..j)};
    od;
    GROp_lcm[p] := convert(map(op, [indices(known)]), set)
  od;

  GROp_r_minus_1 := table();
  for i in [0, r] do
    for q in MSF[i]['factors'] do
      qq := q;
      to r - 1 do
        qq := GOp(qq, t, b); 
        qq := factors(qq);
        qq := op([2, 1, 1], qq); #-
      od;
      GROp_r_minus_1[q] := qq
    od
  od;

  Q_i_set := {}; 
  Q_o_set := {}; 
  F_table := table();
  for q in MSF[r]['factors'] do
    j := 0;
    for p in MSF[0]['factors'] do
      if member(GROp_r_minus_1[q], GROp_lcm[p]) then
        j := j + 1;
        F_table[q, j] := p
      fi;
    od;
    if j > 0 then
      F_table[q] := {seq(F_table[q, jj], jj = 1..j)};
      Q_o_set := Q_o_set union {q}
    else
      Q_i_set := Q_i_set union {q}
    fi;
  od;

  a_check := table();
  for q in MSF[0]['factors'] do
    a_check[q] := MSF[0][q];
    for p in GROp_r_minus_1[q] do
      if normal(MOp(p, t, b)/p/q) = 1 then
        a_check[q] := 0;
        break
      fi
    od
  od;

  b_check := table();
  b_min := table();
  for q in MSF[r]['factors'] do
    b_check[q] := 0;
    for p in GROp_r_minus_1[q] do
      if normal(MOp(p, t, b)/p/q) = 1 then 
        b_check[q] := MSF[r][q];
        break
      fi
    od;
    b_min[q] := b_check[q];
  od;
  a_check[t] := b - 2;
  if assigned(b_check[t]) then
    b_min[t] := max(0, MSF[r][t] - b + 2)
  fi;
  B_start := copy(MSF[r]);
  B_start['coefficient'] := 1;
  for q in MSF[r]['factors'] do
    B_start[q] := b_min[q]
  od;
  B_finish := copy(MSF[r]);
  B_finish['coefficient'] := 1;

  B := copy(B_start);
  while eval(B) <> FAIL do
    deg_B := add(B[q]*degree_table[q], q = MSF[r]['factors']);
    bool := false;
    for p in MSF[r]['factors'] do
      if p <> t and B[p] > 0 then 
        Q := copy(B);
        Q[p] := Q[p] - 1; 
        Q := MSFRatPolyMultiply(msfMp[p], Q, t);
        Q := MSFPolyDivide(MSF[r], Q, t);
        if Q[1] then 
          bool := true;
          break
        fi
      fi;
      if p_divides_Mp[p][1] then 
        Q := MSFRatPolyMultiply(p_divides_Mp[p][2], B, t);
        Q := MSFPolyDivide(MSF[r], Q, t);
        if Q[1] then 
          bool := true;
          break
        fi
      fi
    od;
    if bool then
      B := NextTuple(B, B_start, B_finish, MSF[r]['factors']);
      next
    fi;
    a_bar := table();
    a_max := table();
    for p in MSF[0]['factors'] do
      a_bar[p] := MSF[0][p];
      for q in Q_o_set do
        if B[q] > 0 and member(p, F_table[q]) then
          a_bar[p] := 0;
          break
        fi
      od;
      a_max[p] := min(a_check[p], a_bar[p]);
    od;

    A_start := copy(MSF[0]);
    A_start['coefficient'] := 1;
    for p in MSF[0]['factors'] do
      A_start[p] := 0
    od;
    A_finish := copy(MSF[0]);
    A_finish['coefficient'] := 1;
    for p in MSF[0]['factors'] do
      A_finish[p] := a_max[p]
    od;
    A := copy(A_start);
    while eval(A) <> FAIL do
      deg_A := add(A[p]*degree_table[p], p = MSF[0]['factors']);
      bool := false;
      for p in MSF[0]['factors'] do
        if p <> t then
          Q := MSFPolyDivide(A, msfMp[p], t);
          if Q[1] then 
            Q := Q[2]; 
            Q[p] := Q[p] + 1; 
            Q := MSFPolyDivide(A, Q, t);
            if Q[1] then 
              bool := true;
              break
            fi
          fi
        fi;
        if p_divides_Mp[p][1] then
          Q := MSFPolyDivide(A, p_divides_Mp[p][2], t);
          if Q[1] then 
            bool := true;
            break
          fi
        fi;
      od;
      if bool then
        A := NextTuple(A, A_start, A_finish, MSF[0]['factors']);
        next
      fi;
      polyA := mul(p^A[p], p = MSF[0]['factors']);
      polyB := mul(q^B[q], q = MSF[r]['factors']);
      Ltilde := add(MOp(ell[k], t, br1)
                    * mul(MOp(polyA, t, b^(r-1+j)), j=0..k-1)
                    * mul(MOp(polyB, t, b^j), j=k..r-1) *M^k
                  , k = 0..r);


      for zeta in Zeta_star do
        uae := upper_admissible_edges[zeta];
        nuae := nops(uae);
        uae := [seq(br1*q_L*uae[j][2] 
                  - (br1*deg_A - deg_B)/(b - 1), j = 1..nuae)];
        uae := select(type, uae, nonnegint);
        if uae = [] then
          next
        fi;

        deg_max := uae[-1];

        polyC := LMOpToPolynomialSolutions(
                  eval(Ltilde, M = zeta*M),
                    0, t, M, b,
                    'degree_max' = deg_max,
                    subsop(2 = __C, _options['free'])
                    );


        if polyC <> 0 then
          candidate_counter := candidate_counter + 1;
          candidate_table[candidate_counter] := [zeta, polyA, polyB, polyC];
          if assigned('infolevel'['procname'])
            and 'infolevel'['procname'] > 2 then
            print(candidate_counter = [zeta, polyA, polyB, factor(polyC)])
          fi;
        fi;
      od;
      A := NextTuple(A, A_start, A_finish, MSF[0]['factors']);
    od;
  B := NextTuple(B, B_start, B_finish, MSF[r]['factors'])
  od;


  res :=  map(proc(zABC)
              simplify(
                eval(
                  normal(zABC[1]
                          * MOp(zABC[4], t, b)/zABC[4]
                          * MOp(zABC[2], t, br1)/zABC[3],expanded),
                      t = t^(1/br1)),
                'power', 'symbolic')
             end,
               {seq(candidate_table[i], i=1..candidate_counter)});

  res := RiccatiRatSolCleaning(res, t, __C);
  res := RenamingParameters(res, __C);
  res := eval(res, __C = free);
  res := map(simplify, eval(res, t = x^(1/q_L)), 'power', 'symbolic');
  alg_ext := alg_ext union indets(res, 'RootOf');
  return res
end proc:# ImprovedPetkovsek
#
RiccatiRatSolCleaning := proc(ratset::set(ratpoly(anything, x)),
                                              x, C::{name, symbol})
  local res, r, n, accepted, i, j, ind;
  option procname;
  if has(ratset, RootOf) then
    Normalizer := (evala@Normal)
  fi;
  res := Normalizer(ratset);
  res := [seq([r, select(has, indets(r, name), C)], r = res)];
  res := sort(res,
            proc(r1, r2) evalb(nops(r1[2]) >= nops(r2[2])) end);
  res := [seq(r[1], r = res)];
  n := nops(res);
  accepted := table();
  for i to n do accepted[i] := true od;
  for i from 1 to n - 1 do
    if accepted[i] then
      for j from i + 1 to n do
        if accepted[j] then
          if RationalFunctionsOrientedComparison(
              res[i], res[j], x, C, C) then
            accepted[j] := false
          fi
        fi
      od
    fi
  od;
  ind := [seq(i, i = 1..n)];
  ind := select(proc(ii) evalb(accepted[ii]) end, ind);
  return {seq(res[i], i = ind)}
end proc:# RiccatiRatSolCleaning
#
BasicPetkovsek := proc(L, x, M, b, q_L,
  {field_extension::Or(identical(splitting_field),
                        radalgnum,set(radalgnum)):={},
  free::symbol := `tools/genglobal`('_C')})
  local LL, r, k, ell, br1, candidate_counter, candidate_table,
    MSF_0, MSF_r, factors_list_0, factors_list_r, A_start, p, A_finish,
    B_start, q, B_finish, lin_sys_counter, A, polyA, MpolyA, i, B, polyB,
    BGPform_cond4_not_satisfied, Ltilde, j, upper_admissible_edges,
    Zeta_star, zeta, uae, nuae, Deg_max, polyC, eq, inc, k1, sys, k2,
    sol, res, t::nothing, __C::nothing, alg_ext, tosubs, invtosubs
    ;
  LL := L;

  alg_ext := indets([LL, _options['field_extension']],
    {'radical', 'RootOf'});
  FieldExtension1(alg_ext, 'tosubs', 'invtosubs');
  LL := eval(LL, tosubs);
  if has(LL, RootOf) then
    Normalizer := (evala@Normal);
    LL := collect(LL, M, Normalizer);
  fi;

  LL := eval(L, x = t^q_L);
  LL := primpart(LL, M); 

  r := degree(LL, M);
  for k from 0 to r do
    ell[k] := coeff(LL, M, k)
  od;

  br1 := b^(r-1);
  candidate_counter := 0;
  candidate_table := table();


  MSF_0 := RatPolyToMSF(ell[0], t, _options['field_extension']);
  MSF_r := RatPolyToMSF(ell[r], t, _options['field_extension']);
  factors_list_0 := convert(MSF_0['factors'], list);
  factors_list_r := convert(MSF_r['factors'], list);
  A_start := table([seq(p = 0, p = factors_list_0)]);
  A_finish:= table([seq(p = MSF_0[p], p = factors_list_0)]);
  B_start := table([seq(q = 0, q = factors_list_r)]);
  B_finish:= table([seq(q = MSF_r[q], q = factors_list_r)]);

  lin_sys_counter := 0;
  A := copy(A_start); 
  while eval(A) <> FAIL do
    polyA := mul(p^A[p], p = factors_list_0);
    MpolyA := table();
    for i from 0 to r-1 do
      MpolyA[i] := MOp(polyA, t, b^i);
    od;
    B := copy(B_start);
    while eval(B) <> FAIL do
      polyB := mul(q^B[q], q = factors_list_r);
      BGPform_cond4_not_satisfied := false;
      for i from 0 to r-1 while not BGPform_cond4_not_satisfied do
        if gcd(MpolyA[i], polyB) <> 1 then
          BGPform_cond4_not_satisfied := true;
          break
        fi
      od;
      if BGPform_cond4_not_satisfied then
        B := NextTuple(B, B_start, B_finish, factors_list_r);
        next
      fi;
      Ltilde := add(MOp(ell[k], t, br1)
                    * mul(MOp(polyA, t, b^(r-1+j)), j=0..k-1)
                    * mul(MOp(polyB, t, b^j), j=k..r-1) *M^k
                  , k = 0..r);


      upper_admissible_edges := AdmissibleEdges(Ltilde, t, M, b,
                      'direction' = 'upper',
                      _options['field_extension']);
      Zeta_star := map(op,{indices(upper_admissible_edges)});

      for zeta in Zeta_star do
        uae := upper_admissible_edges[zeta];
        nuae := nops(uae);
        uae := [seq(uae[j][2], j = 1..nuae)];
        uae := select(type, uae, nonnegint);
        if uae = [] then
          next
        fi;
        Deg_max := uae[-1];

        lin_sys_counter := lin_sys_counter + 1;
        polyC := add(__C[k]*t^k, k=0..Deg_max);
        eq := LMOpEval(eval(Ltilde, {M = zeta*M}),
                        polyC, t, M, b);
        inc := {seq(__C[k1], k1 = 0..Deg_max)};
        eq := collect(eq, t);
        sys := {seq(coeff(eq, t, k2), k2 = 0..degree(eq, t))};
        sol := solve(sys, inc);
        if sol <> NULL then
          polyC := eval(polyC, sol);
        else
          next
        fi;

        if polyC <> 0 then
          candidate_counter := candidate_counter + 1;
          candidate_table[candidate_counter] := [zeta, polyA, polyB, polyC];
          if assigned('infolevel'['procname'])
            and 'infolevel'['procname'] > 2 then
            print(candidate_counter = [zeta, polyA, polyB, factor(polyC)])
          fi;
        fi;
      od;  
      B := NextTuple(B, B_start, B_finish, factors_list_r)
    od; 
    A := NextTuple(A, A_start, A_finish, factors_list_0)
  od;  
  res :=  map(proc(zABC)
              simplify(
                eval(
                  normal(zABC[1]
                          * MOp(zABC[4], t, b)/zABC[4]
                          * MOp(zABC[2], t, br1)/zABC[3],expanded),
                      t = t^(1/br1)),
                'power', 'symbolic')
             end,
               {seq(candidate_table[i], i=1..candidate_counter)});

  res := RiccatiRatSolCleaning(res, t, __C);
  res := RenamingParameters(res, __C);
  res := eval(res, __C = free);
  res := map(simplify, eval(res, t = x^(1/q_L)), 'power', 'symbolic');
  return res
end proc:# BasicPetkovsek
#
# Gosper Petkovšek form in the Mahler operator context
GosperPetkovsekForm := proc(u::ratpoly(anything,x), x::name,
                        b::radix, s::nonnegint, $)
  local  U, A, B, alpha, beta, BB, C, G, i, k;
  U := normal(u);
  A := numer(U);
  B := denom(U);
  if A = 0 then
    return 1, 0, 1, 1
  fi;
  alpha := lcoeff(A, x);
  beta := lcoeff(B, x);
  A := A/alpha;
  B := B/beta;
  C := 1;
  for k from 1 to s do
    BB := MOp(B, x, b);
    G := gcd(A, BB);
    A := quo(A, G, x);
    B := quo(BB, G, x);
    C := MOp(C, x, b) * mul(MOp(G, x, b^i),i = 0..k-1)
  od;
  return alpha/beta, A, B, C
end proc:# GosperPetkovsekForm
#
AdmissibleEdges := proc(L, x, M, b,
        {direction::{identical(lower), identical(upper)}:='lower',
        field_extension::Or(identical(splitting_field),
                            radalgnum,set(radalgnum)):={}})
  option procname;
  local edges_list, edges_table, j,
        char_poly, X, obj, i, alpha, z::nothing, fact;

  edges_list := LMOpConvexHull(L, x, M, b, _options['direction'],
    'variable' = 'z');
  edges_table := table();
  for j to nops(edges_list) do
    char_poly := edges_list[j]:-charpoly;
    if 'field_extension' = 'splitting_field' then
      fact := evala(AFactors(char_poly));
    else
      fact := evala(Factors(char_poly, field_extension));
    fi;
    X := [seq(obj[1], obj = op(2,fact))];
    for i to nops(X) do
      if degree(X[i], z) <> 1 then next fi;
      alpha := - coeff(X[i], z, 0)/coeff(X[i], z, 1);
      if assigned(edges_table[alpha]) then
        edges_table[alpha] :=
          [op(edges_table[alpha]),
            [j, - edges_list[j]:-slope, edges_list[j]:-intercept]]
      else
        edges_table[alpha] :=
            [[j, - edges_list[j]:-slope, edges_list[j]:-intercept]]
      fi;
    od;
  od;

  return eval(edges_table)
end proc:# AdmissibleEdges
#
RiccatiByHermitePade := proc(
  L, x, M, b,
    {
    field_extension::Or(identical(splitting_field),
                        radalgnum,set(radalgnum)):={},
    sigma_min::Or(posint,identical(undefined)):='undefined',
    sigma_ratio::Or(realcons,identical(golden_ratio)):='golden_ratio',
    sigma_shift::Or(posint,identical(undefined)):='undefined',
    sigma_max::Or(posint,identical(infinity)):='infinity',
    free::symbol := `tools/genglobal`('_C'),
    prime_decomposition::Or(identical(Maple),identical(Singular)):='Singular'
    })
  local LL, r, k, res, SolBasis, sigma_multiplier, sigma_adder,
    result, lambda, q, alpha, F, ancillary, t_dim, tt_dim, deg_max, br1,
    bound_B_num, bound_B_den, nu, sigma, next_sigma,
    W, deg_W, rho, candidate_set, j, Delta, jj, __C_seq, i, WP,
    choice, id_gen, Sigma, PD, counter, P, genset, gen, sol, K, c, u,
    history, sigma_vect, max_deg_vect, f, slope, intercept, t::nothing,
    y::nothing, ww::nothing, tried_sigma, __C::nothing,
    MF, F_MF, alg_ext, new_alg_ext,
    opt_field_ext, tosubs, invtosubs, newtosubs, newinvtosubs,
    phi := 1.61803398874989484820 
    ;
  
  uses LinearAlgebra;
  LL := L;

  alg_ext := indets([LL, _options['field_extension']], {'radical', 'RootOf'});
  FieldExtension1(alg_ext, 'tosubs', 'invtosubs');
  LL := eval(LL, tosubs);
  alg_ext := eval(alg_ext, tosubs);
  if field_extension = 'splitting_field' then
    opt_field_ext := _options['field_extension']
  else
    opt_field_ext := subsop(2 = alg_ext,
      _options['field_extension']);
  fi;
  if has(LL, RootOf) then
    Normalizer := (evala@Normal);
    LL := collect(LL, M, Normalizer);
  fi;

  r := degree(LL, M);
  if sigma_shift = 'undefined' then
    if type(sigma_ratio, 'realcons') then
     if member(sigma_ratio, {+infinity, -infinity}) then
       error "the option sigma_ratio must provide a real number"
     fi;
     sigma_multiplier := evalf(sigma_ratio, 10);
     if sigma_multiplier <= 1. then
       error "the option sigma_ratio must provide a real number > 1"
     fi;
    elif sigma_ratio = 'golden_ratio' then
     sigma_multiplier := phi
    else
     error "option sigma_ratio not well defined"
    fi;
  else
    if sigma_max = infinity then
     error "sigma_max should have an integer value if sigma_shift is used"
    fi;
    sigma_adder := sigma_shift
  fi;
  Order := 10;
  SolBasis := LMOpToPuiseuxSeriesSolutions(LL, 0, x, M, b,
           opt_field_ext,  
           'output_format' = 'basis', 'ancillary_eqs' = true);
  alg_ext := alg_ext union indets(SolBasis['roots'], {'radical', 'RootOf'});
  FieldExtension1(alg_ext, 'newtosubs', 'newinvtosubs');
  tosubs := tosubs union newtosubs;
  invtosubs := invtosubs union newinvtosubs;
  alg_ext := eval(alg_ext, tosubs);
  alg_ext := remove(type, alg_ext, rational);
  result := table();
  for lambda in SolBasis['roots'] do
    q, F, nu, ancillary := op(SolBasis[lambda]);
    t_dim := nops(F);
    tt_dim := 2*t_dim;
    if t_dim = 0 then
      
      result[lambda] := {};
      next lambda
    fi;
    alpha := SeriesValuation(F[1], x);
    ancillary :=
          LMOpChange(ancillary,
              {x = x, y(x) = x^(+alpha)*ww(x)}, x, M, b);
    deg_max := max(seq(degree(coeff(ancillary, M, k), x), k = 0..r));
    if b = 2 then
      bound_B_num := 2*deg_max;
      bound_B_den := floor(2*(1 - 1/2^r)*deg_max);
    else
      br1 := b^(r-1);
      bound_B_num := floor(4*deg_max/br1);
      bound_B_den := floor(3*deg_max/br1);
    fi;
    result[lambda] := {};
    sigma := max(bound_B_num+1, `if`(sigma_min <> 'undefined', sigma_min, 0));
    history := table();
    for k from 0 do
      if k > 10 and sigma_shift = 'undefined'
        and field_extension <> 'splitting_field' then
        return FAIL
      fi;
      if sigma_shift = 'undefined' then
        if k > 3 and history[k - 1][2] > bound_B_num/5 then
            sigma_vect := Vector([seq(history[i][1], i = k - 3..k - 1)]);
            max_deg_vect := Vector([seq(history[i][2], i = k - 3..k - 1)]);
            f := Statistics:-Fit('intercept' + 'slope'*t,
                  sigma_vect, max_deg_vect, t);
            sol := solve(f = bound_B_num + 10, t);
            if type(sol, {'rational', 'float'}) then
              tried_sigma := ceil(sol)
            fi;
            if type(tried_sigma, 'posint') and tried_sigma > sigma then
              sigma := tried_sigma
            fi;
        fi;
        next_sigma := floor(sigma * sigma_multiplier)
      else
        next_sigma := sigma + sigma_adder
      fi;
      if sigma > sigma_max then
        return FAIL
      fi;

      F := LMOpToLaurentSeriesSolutions(
              ancillary, 0, x, M, b,
              subsop(2 = __C, _options['free']),
              'degree_min' = 0, 'output_format' = 'basis',
              'degree_max' = sigma
              );

      MF := map(MOp, F, x, b);
      F_MF := [op(F), op(MF)];
      W := HermitePadeApproximants(F_MF, x, sigma + 1);
      if W = FAIL then
        return FAIL
      fi;
      W := map(evala, W);

      deg_W := map(degree, W, x);
      history[k] := [sigma, max(op(convert(deg_W, list)))];
      W := SubMatrix(W,
        select( proc(i) local j;
                  evalb(max(seq(deg_W[i, j], j = 1..tt_dim))
                          <= bound_B_num)
                end,[$1..tt_dim]), [$1..tt_dim]);
      rho := RowDimension(W);

      if rho = 0 then
        candidate_set := {};
        break
      fi;
      if rho = tt_dim then
        candidate_set := {};
        sigma := next_sigma;
        next
      fi;
      if rho = tt_dim - 1 then
        for j to tt_dim do
          Delta[j] :=
            Determinant(SubMatrix(W, 1..rho,
              [seq(jj, jj = 1..j-1), seq(jj, jj = j+1..tt_dim)]))
        od;
        candidate_set := {normal(
          add(__C[j]*(-1)^(j+t_dim+1)*Delta[j + t_dim], j = 1..t_dim)/
          add(__C[j]*(-1)^(j      +1)*Delta[j        ], j = 1..t_dim)
          )
        };
      else
        __C_seq := seq(__C[i], i = 1..t_dim);
        WP := < W,
                `<|>`(__C_seq, 0$t_dim),
               `<|>`(0$t_dim, __C_seq)>;
        choice := combinat:-choose(tt_dim, rho + 2);
        for i to binomial(tt_dim, rho + 2) do
          Delta[i] :=
            Determinant(SubMatrix(WP, 1..rho+2, choice[i]));
          Delta[i] := {coeffs(collect(Delta[i], x), x)};
        end do;
        id_gen := `union`(seq(Delta[i], i = 1..binomial(tt_dim, rho + 2)));

        Sigma := PolynomialIdeals:-PolynomialIdeal(id_gen,
          'characteristic' = 0, 'variables' = {__C_seq});
        local singular_tim := -1, singular_ver := -1;
        if field_extension = 'splitting_field' then
          PD, new_alg_ext, singular_tim, singular_ver :=
            AbsolutePrimeDecomposition(Sigma, _options['prime_decomposition']);
          alg_ext := alg_ext union new_alg_ext;
        else
          PD := [PolynomialIdeals:-PrimeDecomposition(Sigma,
            alg_ext, 
            'removeredundant' = true)];
        fi;
        res := table();
        counter := 0;
        for j to nops(PD) do
          P := PD[j];
          genset := PolynomialIdeals:-Generators(P);
          for gen in genset do
            if not (degree(gen, {__C_seq}) = 1) then
              sigma := next_sigma;
              next k
            fi
          end do;

          sol := solve(genset, {__C_seq});
          if eval({__C_seq}, sol) = {0} then
            next j
          fi;
          K := NullSpace(eval(Transpose( WP), sol));
          if nops(K) > 1 then
            sigma := next_sigma;
            next k
          fi;
          K := op(K);
          if K[rho + 1] = 0 or K[rho + 2] = 0 then
            sigma := next_sigma;
            next k
          fi;
          counter := counter + 1;
          res[counter] := Normalizer(- K[rho + 1]/K[rho + 2]);
        end do; 
        candidate_set := {seq(res[c], c = 1..counter)};

      fi;

      candidate_set := candidate_set minus result[lambda];
      for u in candidate_set do
        if degree(numer(u), x) > bound_B_num
          or degree(denom(u), x) > bound_B_den then
          sigma := next_sigma;
          next k
        fi;
        if evala(LMOpHornerScheme(ancillary, u, x, M, b)) = 0 then
          result[lambda] := result[lambda] union {u};
        else
          sigma := next_sigma;
          next k
        fi;
      end do; 
      break
    od; 
    result[lambda] := RiccatiRatSolCleaning(result[lambda], x, __C);
    result[lambda] :=
      map(foo -> lambda*eval(x^((b-1)*alpha)*foo, x = x^(1/q)),
          result[lambda]);
  od; 

  res := `union`(seq(result[lambda], lambda = SolBasis['roots']));
  res := RenamingParameters(res, __C);
  res := eval(res, __C = free);
  res := eval(res, invtosubs);
  return res
end proc:# RiccatiByHermitePade
#
AbsolutePrimeDecomposition := proc(Sigma, {
  prime_decomposition::Or(identical(Maple),identical(Singular)):='Singular'
  })
  if prime_decomposition = 'Maple' then
    return HeurAbsPrimeDecomp(Sigma), -1, -1
  else
    local resp :=  ssystem("which Singular");
    if resp[1] = 1 then
      WARNING("Singular not available, trying an incomplete heuristic");
      return HeurAbsPrimeDecomp(Sigma), -1, -1;
    elif resp[1] <> 0 then error "unexpected result after testing Singular" fi;
    local apd, tim, ver;
    apd, tim, ver := AbsolutePrimeDecompositionBySingular(Sigma);
    return map2(op, 1, apd), indets(apd, 'RootOf'), tim, ver
  fi
end proc:# AbsolutePrimeDecomposition
#
HeurAbsPrimeDecomp := proc(Sigma)
local old_ext, new_ext, other::nothing, PD, genprod, fact;
  old_ext := other;
  new_ext := indets(Sigma, 'RootOf');
  while old_ext <> new_ext do
    PD := [PolynomialIdeals:-PrimeDecomposition(Sigma, new_ext,
      'removeredundant' = true)];
    genprod := convert(convert(map(op, map(PolynomialIdeals:-Generators, PD)),
      set), `*`);
    fact := evala(AFactors(genprod));
    old_ext := new_ext;
    new_ext := indets(fact, 'RootOf');
  od;
  return PD, old_ext
end proc:# HeurAbsPrimeDecomp
#
HermitePadeApproximants := proc(FF::list, z::name, NN)
    local i, d, n, vn, F, M, N;
    option procname;
    if has(FF, 'RootOf') then
      WARNING("series contain RootOf's so solutions could be lost")
    fi;
    n := nops(FF);
    if _npassed = 4 then
        N := NN;
        vn := _passed[4];
    elif _npassed = 3 and type(NN, list) then
        vn := NN;
        N := add(NN[i], i = 1 .. n) + n - 1;
    else
        N := NN;
        vn := [seq(N, i = 1 .. n)];
    end if;
    F := map(convert, map(series, FF, z, N + 1), polynom);  
    M := MatrixPolynomialAlgebra:-MahlerSystem(Matrix(1, n, F), z, vn, [N]);
    M := map(evala, M);
    d := lcoeff(M[1, 1], z);
    if d = 0 then return FAIL fi;
    M := map(x -> normal(x/d), M);
    LinearAlgebra:-Transpose(M);
end proc:# HermitePadeApproximants
#
# definition of a Mahlerian hypergeometric function
MHypergeom := proc(g::algebraic, x, b::radix,
      {field_extension::Or(radalgnum,set(radalgnum)):={}})
  local q, Zeta, tosubs, invtosubs, gg, ss, lambda, m, ggg, h, f;
  option procname;
  q := RamificationOrder(g, x);
  if q <> FAIL then
    gg := simplify(eval(g, x = x^q), 'power', 'symbolic');
    Zeta := FieldExtension1(field_extension, 'tosubs', 'invtosubs');
    if Zeta <> {} and has(Zeta, RootOf) then
      Normalizer := (evala@Normal)
    fi;
    gg := Normalizer(gg);
    if gg <> 0 then
      ss := MultiSeries:-multiseries(gg, x, 2);
      lambda := MultiSeries:-SeriesInfo:-ListCoefficients(ss)[1];
      m := MultiSeries:-SeriesInfo:-ListExponents(ss)[1];
      ggg := Normalizer(gg/lambda/x^m);
      h, f := EvanescentProductCleaning(ggg, x, b,
               subsop(2 = Zeta, _options['field_extension']));
      if h = 1 then
        return ln(1/x)^log[b](lambda) *x^(m/q/(b - 1)) *
               rationalize(simplify(eval(f, x = x^(1/q)), 'power', 'symbolic'))
      else
      return ln(1/x)^log[b](lambda) *x^(m/q/(b - 1)) *
        rationalize(simplify(eval(f, x = x^(1/q)), 'power', 'symbolic')) *
        'MHypergeom'(
          rationalize(simplify(eval(h, x = x^(1/q)), 'power', 'symbolic')),
                 x, b)
      fi
    else
      return NULL
    fi
  else
    error cat("a ramified rational function in %2 is expected, ",
                "but %1 is received"), g, x
  fi;
end proc:# MHypergeom
#
LMOpToMHypergeometricSolutions := proc(L, x, M, b,
  {output_format::Or(identical(undefined),identical(basis)):='undefined',
  free::symbol := `tools/genglobal`('_C'),
  field_extension::Or(identical(splitting_field),
                      radalgnum,set(radalgnum)):={}},
  $)
  option procname;
  local result, res, c, i, res_i, nops_i, j, U, u, ind, uu, cc;

  if output_format = 'undefined' then
    result := LMOpToMHypergeometricSolutions(L, x, M, b,
                  subsop(2 = 'basis', _options['output_format']),
                  _options['field_extension'],
                  _options['free']);
    res := table();
    c := 0;
    for i to nops(result) do
      res_i := op(i, result);
      nops_i := nops(res_i);
      res[i] := factor(add(free[c + j]*op(j, res_i), j = 1..nops_i));
      c := c + nops_i;
    od;
    return {seq(res[i], i = 1..nops(result))}
  fi;
  U := RiccatiSolve(_params[1.._nparams],
                    _options['field_extension'],
                    _options['free']);
  U := U minus {0};
  res := table();
  for u in U do
    ind := indets(u, name) minus {x};
    ind := select(has, ind, free);
    if ind = {} then
      res[u] := [MHypergeom(u, x, b)]
    else
      for c in ind do
        uu[c] := eval(u, {c = 1, seq(cc = 0, cc = ind minus {c})})
      od;
      res[u] := map(MHypergeom, [seq(uu[c], c = ind)], x, b);
      res[u] := sort(res[u],
                  `=`('key',sp->SeriesValuation(
                                eval(sp, {MHypergeom = 1, ln = 1}), x)));
    fi
  od;
  return {seq(res[u], u = U)}
end proc:# LMOpToMHypergeometricSolutions
#
EvanescentProductCleaning := proc(g::ratpoly(anything, x), x, b,
    {field_extension::set(radalgnum):={}})
  local MSF, gg, f, nonstop, j, SB, ff, i, p, q, nu, bool, s, minn;
  gg := g;
  f := 1;
  nonstop := true;
  for j while nonstop do
    if assigned('infolevel'['procname'])
        and 'infolevel'['procname'] > 1 then
      printf("loop number %d\n", j);
    fi;
    MSF := RatPolyToMSF(gg, x,
          _options['field_extension'], 'direction' = 'lower');
    MSF['factors'] := MSF['factors']  minus {x};
    SB[0], SB[1] := MSFToSiblingsGroups(MSF, x, b,
          _options['field_extension'], 'direction' = 'lower');
    ff := 1;
    for i from 0 to 1 do
      for p in SB[i][factors] do
        if nops(SB[i][p]) = 1 then
          q := op(1, SB[i][p]);
          nu := MSF[q];
          ff := p^nu
        else
          bool := true;
          s := sign(MSF[op(1, SB[i][p])]);
          minn := +infinity;
          for q in SB[i][p] do
            if s*sign(MSF[q]) > 0 then
              minn := min(abs(MSF[q]), minn);
            else
              bool := false;
              break
            fi;
          od;
          if bool then
            nu := s*minn;
            ff := p^nu
          fi;
        fi;
        gg := normal(gg*(ff/MOp(ff, x, b)));
        f := f*ff
      od;
    od;
    if assigned('infolevel'['procname'])
        and 'infolevel'['procname'] > 1 then
      printf("new g = %s, newf = %s\n",
          convert(gg, string), convert(f, string));
    fi;
    if ff = 1 then nonstop := false fi;
  od;

  return gg, f
end proc:# EvanescentProductCleaning
cachetableEvanescentProductCleaning:= Cache(100):
Cache(cachetableEvanescentProductCleaning,
  procedure = EvanescentProductCleaning):
#
MSFToSiblingsGroups := proc(MSF, x, b,
      {field_extension::set(radalgnum):={},
        direction::{identical(lower), identical(upper)}:='upper'})
  option procname;
  local to_deal_with, q, p,
    SB, bs, sg_bool, i, father, siblings ;
  to_deal_with := MSF['factors'];
  SB[0] := table();
  SB[1] := table();

  for q in to_deal_with do
    p := GOp(q, x, b);
    father := RatPolyToMSF(p, x, b, _options); 
    p := op(father['factors']);
    siblings := RatPolyToMSF(MOp(p, x, b),
                              x, b, _options['direction']);
    siblings := siblings['factors'];
    if member(p, siblings) then
      SB[0][p] := siblings minus {p}
    else
      SB[1][p] := siblings
    fi;
    to_deal_with := to_deal_with minus siblings
  od;
  for i from 0 to 1 do
    for p in op~({indices(SB[i])}) do
      sg_bool := true; 
      for bs in SB[i][p] do 
        if not member(bs, MSF['factors']) then
          sg_bool := false;
          break
        fi
      od;
      if not sg_bool then
        unassign(evaln(SB[i][p]))
      fi
    od;
    SB[i]['factors'] := op~({indices(SB[i])})
  od;

  return eval(SB[0]), eval(SB[1])
end proc:# MSFToSiblingsGroups
#
IQuo := proc(a::{integer, infinity, -infinity},b::posint,r::name)
  local q, rr;
  option procname;
  q := iquo(a,b,'rr');
  if type(a, 'negint') and rr <> 0 then
    q := -1 + q;
    rr := b + rr
  fi;
  if _npassed > 2 then
    r := rr
  fi;
  return q
end proc:# IQuo
#
LMOpConvexHull := proc(L::polynom(anything,{x,M}), x::name, M::name, b::radix,
        {direction::{identical(lower), identical(upper),
                      identical(undefined)}:='undefined',
         output::{identical(geometry), identical(charpolys)}:='charpolys',
         variable::name:='lambda'})
  local LL, r, k, ell, deg, degg, CH, rec, data, c;
  if direction = 'undefined' then
    return  LMOpConvexHull(L, x, M, b,
              subsop(2 = 'lower', _options['direction']),
              _options['output'], _options['variable']),
            LMOpConvexHull(L, x, M, b, lambda,
              subsop(2 = 'upper', _options['direction']),
              _options['output'], _options['variable'])
  fi;

  LL := collect(L, M, expand);
  if LL = 0 then return [] fi;

  r := degree(LL, M);
  for k from 0 to r do
    ell[k] := expand(coeff(LL, M, k));
    deg[k] := degree(ell[k], x);
  od;

  if 'direction' = 'upper' then
    degg := max(seq(deg[k], k=0..r));
    LL := collect(numer(normal(eval(LL, {x = 1/x}))), M, expand);
    CH := LMOpConvexHull(LL, x, M, b,
            subsop(2 = 'lower', _options['direction']),
            _options['output'], _options['variable']);
    if output = 'geometry' then
      return map(pt->[pt[1], degg - pt[2]], CH)
    else
      return [seq(eval(Record(
          :-abscissa = rec:-abscissa,
          :-ordinate = degg - rec:-ordinate,
          :-slope = - rec:-slope,
          :-intercept = degg - rec:-intercept,
          :-charpoly = rec:-charpoly
          )), rec = CH)]
    fi;
  fi;

  data := table();
  c := 0;
  for k from 0 to r do
    if ell[k] <> 0 then
    c := c + 1;
    data[c] := [k, [b^k, ldegree(ell[k], x)], tcoeff(ell[k], x)]
    fi
  od:
  
  CH := JustDoLMOpConvexHull(data);
  if 'output' = 'geometry' then
    CH := GeomConvexHull(CH)
  else
    CH := CharPolysConvexHull(CH, op(2, _options['variable']))
  fi;
  return eval(CH);
end proc:# LMOpConvexHull
cachetableLMOpConvexHull  := Cache(100):
Cache(cachetableLMOpConvexHull, procedure=LMOpConvexHull):
#
JustDoLMOpConvexHull := proc(data)
  local c, lch, clch, i, pt;
  c := max(map(op, ([indices(data)])));
  lch := table();
  if c = 1 then
    return table([1 = data[1]])
  fi;
  clch := 0;
  for i to c do
    pt := data[i][2];
    while (clch >= 2 and Cross(lch[clch - 1][2], lch[clch][2],pt) < 0) do
      clch := clch - 1
    od;
    clch := clch + 1;
    lch[clch] := data[i];
  od;
  return eval(table([seq(i = lch[i], i = 1..clch)]))
end proc:# JustDoLMOpConvexHull
#
CharPolysConvexHull := proc(lch::table, lambda::name)
  local clch, vic, viccounter, abscissa, ordinate, slope, intercept,
    e0, tc, charpoly, i, e, newabscissa, newordinate, newslope, newintercept;
  clch := max(op(map(op, [indices(lch)])));
  vic := table();
  viccounter := 0;
  abscissa := lch[1][2][1];
  ordinate := lch[1][2][2];
  if clch = 1 then 
    return [Record(
        :-abscissa = abscissa,
        :-ordinate = ordinate,
        :-slope = +infinity,
        :-intercept = -infinity,
        :-charpoly = 0
        )] 
  fi;
  slope := Slope(lch[1][2], lch[2][2]);
  intercept := Intercept(lch[1][2], lch[2][2]);
  e0 := lch[1][1];
  tc := lch[1][3];
  charpoly := tc;
  i := 2;
  while i < clch do
    e := lch[i][1];
    tc := lch[i][3];
    if e0 = -infinity then
      charpoly := 1 - lambda 
    else
      charpoly := charpoly + tc * lambda^(e - e0)
    fi;
    newabscissa := lch[i][2][1];
    newordinate := lch[i][2][2];
    newslope :=
      Slope(lch[i][2], lch[i+1][2]);
    newintercept :=
      Intercept(lch[i][2], lch[i+1][2]);
    if newslope <> slope then
      viccounter := viccounter + 1;
      vic[viccounter]
        := Record(:-abscissa = abscissa, :-ordinate = ordinate,
      :-slope = slope, :-intercept = intercept, :-charpoly = charpoly);
      abscissa := newabscissa;
      ordinate := newordinate;
      slope := newslope;
      intercept := newintercept;
      charpoly := tc;
      e0 := lch[i][1];
    fi;
    i := i + 1;
  od;
  e := lch[i][1];
  tc := lch[i][3];
  if e0 = -infinity then
    charpoly := 1 - lambda  
  else
    charpoly := charpoly + tc * lambda^(e - e0)
  fi;
  charpoly := primpart(charpoly, lambda);
  viccounter := viccounter + 1;
  vic[viccounter]
    := Record(:-abscissa = abscissa, :-ordinate = ordinate,
      :-slope = slope, :-intercept = intercept, :-charpoly = charpoly);
  viccounter := viccounter + 1;
  abscissa := lch[i][2][1];
  ordinate := lch[i][2][2];
  slope := +infinity;
  intercept := -infinity;
  charpoly := 0;
  vic[viccounter]
    := Record(:-abscissa = abscissa, :-ordinate = ordinate,
      :-slope = slope, :-intercept = intercept, :-charpoly = charpoly);
  return [seq(eval(vic[i]), i=1..viccounter)]
end proc:# CharPolysConvexHull
#
GeomConvexHull := proc(lch::table)
local c, i;
  c := max(map(op, [indices(lch)]));
  return [seq(lch[i][2], i = 1..c)]
end proc:# GeomConvexHull
#
Cross := proc(P::list,A::list,B::list)
  option procname;
 (A[1] - P[1])*(B[2] - P[2]) - (A[2] - P[2])*(B[1] - P[1])
end proc:# Cross
#
ExtendedInverse := proc(v)
  option procname;
  if v = 0 then
    infinity
  elif v = infinity then
    0
  else
    1/v
  fi
end proc:# ExtendedInverse
#
Slope := proc(A::list,B::list)
  option procname;
  (B[2] - A[2])* ExtendedInverse(B[1] - A[1])
end proc:# Slope
#
Intercept := proc(A::list,B::list)
  option procname;
  (A[2]*B[1] - A[1]*B[2])* ExtendedInverse(B[1] - A[1])
end proc:# Intercept
#
SeriesValuation := proc(f::algebraic, x)
  local ff, q, s, msg;
    
  if type(f, 'series') then
    
    if op(0, f) <> x then
      error "%1 is assumed to be a series in %s\n", f, convert(x,string)
    fi;
    return op(2, f)
  elif type(f, 'puiseuxseries') then
    
    q := op(1, f);
    s := op(2, f);
    return SeriesValuation(s, x)/q
  elif type(f, 'ratpoly'('anything', x)) then
    
    ff := normal(f);
    if ff = 0 then
      return +infinity
    fi;
    ff := MultiSeries:-multiseries(ff, x, 3);
    return MultiSeries:-SeriesInfo['ListExponents'](ff)[1];
  elif type(f, 'ramratpoly'('anything', x)) then
    
    q := RamificationOrder(f, x);
    ff := simplify(eval(f, x = x^q), 'power', 'symbolic');
    return SeriesValuation(ff, x)/q
  else
    msg := cat("%1 expect its first argument to be a series or ",
      "a Puiseux series or rational function or a ramified rational ",
      "function, but %2 is received");
    error msg, procname, f
  fi
end proc:# SeriesValuation
#
NextTuple := proc(r::table, start::table, finish::table,
                  k::{nonnegint, list, set})
  local kk, c, i;
  option procname;
  if type(k, 'nonnegint') then
    kk := [$k];
  else
    kk := k
  fi;
  for i in kk do
    if start[i] > finish[i] then return FAIL fi
  od;
  c := 1;
  for i in kk do
    r[i] := r[i] + c;
    if r[i] > finish[i] then r[i] := start[i]; c := 1
    else c:= 0; break
    fi;
  od;
  if c = 1 then return FAIL else return eval(r) fi
end proc:# NextTuple
#
RatPolyToMSF := proc(f::ratpoly(anything, x), x::name,
    {field_extension::Or(identical(splitting_field),
                        radalgnum,set(radalgnum)):={},
    direction::{identical(lower), identical(upper)}:='upper'})
  option remember;
  option procname;
  local factzdf, ff, num, den, MSFnum, MSFden, koeff, k, i, T, newT;
  if 'direction' = 'upper' then
    if field_extension = 'splitting_field' then
      if type(f, 'polynom') then
        factzdf := evala(AFactors(f)) 
      else  
        ff := normal(f);
        num := numer(ff); den := denom(ff);
        MSFnum := RatPolyToMSF(num, x,
          _options['direction'], _options['field_extension']);
        MSFden := RatPolyToMSF(den, x,
          _options['direction'], _options['field_extension']);
        return MSFRatPolyMultiply(MSFnum, MSFden, x, 'exponents' = [1, -1])
      fi
    else
      factzdf := factors(f, field_extension); 
    fi;
    T := table();
    T[mon_order] := direction;
    T['coefficient'] := op(1, factzdf);
    factzdf := op(2, factzdf);
    koeff := remove(has, factzdf, x);
    factzdf := select(has, factzdf, x); 
    T['coefficient'] := T['coefficient']*mul(k[1]^k[2], k = koeff);
    T['factors'] := convert(map2(op, 1, factzdf),set);
    for i to nops(factzdf) do
      T[factzdf[i][1]] := factzdf[i][2]
    od;
    return eval(T)
  else
    T := RatPolyToMSF(f, x, 'direction' = 'upper', _options['field_extension']);
    newT := MSFToMSF(T, x, _options['direction']);
    return eval(newT)
  fi
end proc:# RatPolyToMSF
#
MSFToMSF := proc(T, x,
    {direction::{identical(lower), identical(upper)}:= NULL})
  local leadingcoeff, Link, lc, p, newT;
  option procname;
  if direction = NULL or direction = T['mon_order'] then
    return eval(T)
  fi;
  if 'direction' = 'upper' then
    leadingcoeff := eval(lcoeff)
  else
    leadingcoeff := eval(tcoeff)
  fi;
  Link := table();
  lc := T['coefficient'];
  for p in T['factors'] do
    Link[p] := expand(p/leadingcoeff(p, x));
    lc := lc*leadingcoeff(p, x)^T[p]
  od;
  newT := table();
  newT['mon_order'] := direction;
  newT['coefficient'] := lc;
  newT['factors'] := {seq(Link[p], p = T['factors'])};
  for p in T['factors'] do
    newT[Link[p]] := T[p]
  od;
  return eval(newT)
end proc:# MSFToMSF
#
MSFRatPolyMultiply := proc(T1, T2,  x,
    {direction::{identical(lower), identical(upper),identical(NULL)}:= NULL,
      exponents::{list(integer),identical(NULL)}:=NULL})
  local PI, p, newT1, newT2, e1, e2, msg;
  option procname;
  if direction = NULL then
    return MSFRatPolyMultiply(T1, T2, x,
      subsop(2 = T1['mon_order'], _options['direction']),
      _options['exponents'])
  else
    newT1 := MSFToMSF(T1, x, _options['direction']);
    newT2 := MSFToMSF(T2, x, _options['direction']);
  fi;
  if exponents = NULL then
    e1 := 1; e2 := 1;
  else
    if nops(exponents) <> 2 then
      msg := cat("the keyword parameter exponents is assumed to provide ",
                  "a list of two integers, but %1 is received");
      error msg, exponents
    fi;
    e1 := exponents[1]; e2 := exponents[2];
  fi;
  PI := table();
  for p in newT1['factors'] do
    PI[p] := e1*newT1[p]
  od;
  for p in newT2['factors'] do
    if assigned(PI[p]) then
      PI[p] := PI[p] + e2*newT2[p]
    else
      PI[p] := e2*newT2[p]
    fi
  od;
  PI['factors'] := map(op, {indices(PI)}); 
  PI['mon_order'] := direction;
  PI['coefficient'] := newT1['coefficient']^e1 * newT2['coefficient']^e2;
  return eval(PI)
end proc:# MSFRatPolyMultiply
#
MSFPolyDivide := proc(T1, T2, x,
    {direction::{identical(lower), identical(upper), identical(NULL)}:= NULL})
  local Q, p;
  option procname;
  Q := MSFRatPolyMultiply(T1, T2, x, _options['direction'],
          'exponents' = [1, -1]);
  for p in Q['factors'] do
    if Q[p] < 0 then return [false] fi
  od;
  return [true, eval(Q)]
end proc:# MSFPolyDivide
#
WellChosenDenominator := proc(ratlist, b::posint)
  option procname;
  local denomlist;
  denomlist := map(denom, ratlist);
  denomlist := select(proc(foo) evalb(igcd(foo, b) = 1) end, denomlist);
  return ilcm(op(denomlist))
end proc:# WellChosenDenominator
#
RamificationOrder := proc(f::{algebraic, set(algebraic)},
  v::{name, set(name), list(name)}, $)
local w, msg, q, ff, x;
  if type(v, 'list'('name')) then
    w := v
  elif type(v, 'set'('name')) then
    w := convert(v, list)
  elif type(v, 'name') then
    w := [v]
  else
    msg := cat("%1 expects its second argument to be of type ",
               "name or set(name) or list(name), "
               "but %2 is received");
    error msg, procname, v
  fi;
  if has(f, FAIL) then return FAIL fi;
  q := JustDoRamificationOrder(f, w);
  if q = FAIL then return FAIL fi;
  if convert(q, set) = {1} then
    if type(v, 'list'('name')) then
      return [seq(1, x = w)]
    else
      return 1
    fi
  fi;
  ff := subs({seq(x = x^q[x], x = w)}, f);
  ff := simplify(ff, power, symbolic);
  ff := normal(ff, expanded);
  ff := subs({seq(x = x^(1/q[x]), x = w)}, ff);
  ff := simplify(ff, power, symbolic);
  q := JustDoRamificationOrder(ff, w);
  if type(v, 'list'('name')) then
    return [seq(q[x], x = w)]
  else
    return ilcm(seq(q[x], x = w))
  fi
end proc:# RamificationOrder
#
JustDoRamificationOrder := proc(f::{algebraic, set(algebraic)}, w::list(name))
local ff, vv, msg, q_tbl, x;
  option remember;
  if type(f, 'series') or has(indets(f, 'function'), 'O') then
    return FAIL
  fi;
  if type(f, 'ratpoly') then
    ff := f
  else
    ff := simplify(f, 'power', 'symbolic')
  fi;
  vv := indets(ff);
  vv := select(has, vv, w);
  q_tbl := table();
  for x in w do
    if type(ff, 'ratpoly'('anything', x)) then
      q_tbl[x] := 1
    else
      q_tbl[x] := CommonRamificationOrder(select(has, vv, x), {x});
      if q_tbl[x] = FAIL then return FAIL fi
    fi;
  od;
  return eval(q_tbl)
end proc:# JustDoRamificationOrder
#
CommonRamificationOrder := proc(s::set,
    v::Or(list({name, symbol}),set({name, symbol})))
  local ind, exp_tb, j, jj, aa::nothing, tosubs::nothing, z::nothing;
  option procname;
  ind := select(has, s, v) minus convert(v, set);
  if ind = {} then
    return 1
  fi;
  exp_tb := table():
  for j to nops(ind) while
    patmatch(op(j, ind), ('z'::name)^('aa'::rational), 'tosubs') do
      if member(eval(z, tosubs),  v) then
       exp_tb[j] := eval(aa, tosubs)
      else
        return FAIL
      fi
  od;
  if j = nops(ind) + 1 then
    return ilcm(seq(denom(exp_tb[jj]), jj = 1..j))
  else
    return FAIL
  fi;
end proc:# CommonRamificationOrder
#
# structured expression for Puiseux series
PuiseuxSeries := proc(q::posint, s::{series, 0})
  local n, i, r, c, e, x;
  if q = 1 then
    s
  elif s = 0 then
    0
  else
    n := iquo(nops(s), 2);
    for i to n do
      e[i] := iquo(op(2*i, s), q, 'r');
      c[i] := op(2*i - 1, s);
      if r <> 0 then
        break
      fi;
    od;
    if i = n + 1 then
      x := op(0, s);
      series(add(c[i]*x^e[i], i = 1..n), x, e[n])
    else
      'procname'(q, s)
    fi
  fi;
end proc:# PuiseuxSeries
#
CompRatFunSets := proc(
  rs1, rs2,
  x::name, C1::{name, symbol}, C2::{name, symbol})
  local n1, n2, r1, r2, notdealtwith1, notdealtwith2, i1, i2,
    test, q, newrs1, newrs2, msg;
  option procname;
  if has(rs1 union rs2, RootOf) then
    Normalizer := (evala@Normal)
  fi;
  if type(rs1, 'set'('ratpoly'('anything', x)))
    and type(rs2, set(ratpoly(anything, x))) then
    r1 := convert(rs1, list);
    r2 := convert(rs2, list);
    n1 := nops(r1);
    n2 := nops(r2);
    notdealtwith1 := table();
    notdealtwith2 := table();
    for i1 to n1 do notdealtwith1[i1] := true od;
    for i2 to n2 do notdealtwith2[i2] := true od;
    for i1 to n1 do
      if notdealtwith1[i1] then
        for i2 to n2 do
          if notdealtwith2[i2] then
            if ComparingRationalFunctions(r1[i1], r2[i2], x, C1, C2) then
              notdealtwith1[i1] := false;
              notdealtwith2[i2] := false;
              next
            fi;
          fi
        od
      fi;
      if notdealtwith1[i1] then
        return false
      fi
    od;
    for i2 to n2 while not notdealtwith2[i2] do od;
    if i2 > n2 then
      return true
    else
      return false
    fi;
  fi;

  q := RamificationOrder(rs1 union rs2, x);
  if q <> FAIL then
    newrs1 := map(MOp, rs1, x, q);
    newrs2 := map(MOp, rs2, x, q);
    return CompRatFunSets(newrs1, newrs2, x, C1, C2)
  else
    return false
  fi;
end proc:# CompRatFunSets
#
ComparingRationalFunctions := proc(
    r1::ratpoly(anything, x), r2::ratpoly(anything, x),
    x::name, C1::{name, symbol}, C2::{name, symbol})
    option procname;
    evalb(
      RationalFunctionsOrientedComparison(r1, r2, x, C1, C2)
      and
      RationalFunctionsOrientedComparison(r2, r1, x, C2, C1)
    )
end proc:# ComparingRationalFunctions
#
RationalFunctionsOrientedComparison := proc(
    ratf1::ratpoly(anything, x), ratf2::ratpoly(anything, x),
    x::name, C1::{name, symbol}, C2::{name, symbol})
  local K::nothing, rf1, rf2, EQ, k, eq, inc, sol, i;
  option procname;
  rf1 := normal(ratf1, expanded);
  rf2 := normal(ratf2, expanded);
  rf1 := eval(rf1, C1 = K);
  EQ := collect(numer(Normalizer(rf1 - rf2)), x);
  if EQ = 0 then return true fi;
  for k from 0 to degree(EQ, x) do
    eq[k] := coeff(EQ, x, k)
  od;
  inc := select(has, indets(rf1, {name, symbol}), K);
  if inc = {} then return false fi;
  sol := solve({seq(eq[k], k = 0..degree(EQ, x))}, inc);
  if sol <> NULL and sol <> {} and sol <> {seq(i = 0, i = inc)} then
    if _npassed > 5 and type(_passed[6], name) then
      assign(_passed[6], eval(sol, K = C1))
    fi;
    return true
  else
    return false
  fi;
end proc:# RationalFunctionsOrientedComparison
#
RenamingParameters := proc(s::set, K::{name, symbol})
  local N, j, i, ind, k, tosubs, ii, news;
  N := nops(s);
  j := 0;
  for i to N do
    ind := select(has, indets(op(i, s), name), K);
    k := nops(ind);
    tosubs := {seq(op(ii, ind) = K[j + ii], ii = 1..k)};
    news[i] := eval(op(i, s), tosubs);
    j := j + k;
  od;
  return {seq(news[i], i = 1..N)};
end proc:# RenamingParameters
#
FieldExtension1 := proc(field_ext)
  local Zeta, zeta, tosubs, invtosubs, indRootOf, TOSUBS;
  option procname;
  if not member(field_ext, {'ground_field', 'base_field'}) then
    if type(field_ext, 'set') then
      Zeta := field_ext
    elif type(field_ext, 'list') then
      Zeta := convert(field_ext, set)
    else
      Zeta := {field_ext}
    fi
  else
    Zeta := {}
  fi;
  Zeta := remove(type, Zeta, rational);
  tosubs := {seq(zeta = evala(Algebraic:-ConvertRootOf(zeta)), zeta = Zeta)};
  indRootOf := indets(tosubs, 'RootOf');
  TOSUBS := map(obj->obj='RootOf'(op(1, obj)), indRootOf);
  tosubs := eval(tosubs, TOSUBS);
  if _npassed > 2 and type(_passed[2], name) and type(_passed[3], name) then
    invtosubs := map(obj -> op(2, obj) = op(1, obj), tosubs);
    assign(_passed[2], tosubs);
    assign(_passed[3], invtosubs)
  fi;
  return eval(Zeta, tosubs);
end proc:# FieldExtension1
#
FieldExtension2 := proc(polyset, v)
  local Zeta, f, obj, tosubs, invtosubs;
  option procname;
  f := mul(obj, obj = polyset);
  PolynomialTools:-Splits(f, v, 'Zeta');
  Zeta := remove(type, Zeta, rational);
  tosubs := {};
  invtosubs := {};
  if _npassed > 3 and type(_passed[3], name) and type(_passed[4], name) then
    assign(_passed[3], tosubs);
    assign(_passed[4], invtosubs)
  fi;
  return eval(Zeta, tosubs);
end proc:# FieldExtension2
#
# radix type
`type/radix` := proc(b)
  type(b, integer) and evalb(b > 1)
end proc:# `type/radix`
#
# ramified rational function type
`type/ramratpoly` := proc(f, ground_field, v)
local ind, vv, q, ff, i;
  ind := indets(f);
  vv := select(type, ind, 'name');
  if _npassed = 1 then
    return thisproc(f, anything, vv)
  elif _npassed = 2 then
    return thisproc(f, ground_field, vv)
  elif not member(_npassed, {3, 4})  then
    error "wrong type of arguments"
  else
    if type(v, 'set') then
      vv := convert(v, list)
    elif type(v, 'list') then
      vv := v
    else
      vv := [v]    fi
  fi;
  ind := select(has, ind, vv);
  if type(f, 'ratpoly'(ground_field, v)) then
    return true
  else
    q :=  dcfun:-RamificationOrder(f, vv);
    if q = 'FAIL' then return false fi;
    ff := eval(f, {seq(vv[i] = vv[i]^q[i], i = 1..nops(vv))});
    ff := simplify(ff, 'power', 'symbolic');
    return type(ff, ratpoly(ground_field, vv))
  fi;
end proc:# `type/ramratpoly`
#
# formal Laurent series type
`type/laurentseries` := proc(f, ground_field, v)
  local vv, x, ff, omega, i;
  option procname;
  if not type(f, 'algebraic') then
    return false
  fi;
  vv := indets(f, name);
  if vv = {} then return false fi;
  if _npassed = 1 then
    return thisproc(f, anything, vv)
  elif _npassed = 2 then
    return thisproc(f, ground_field, vv)
  elif not _npassed = 3  then
    error "wrong type of arguments"
  else
    if type(v, 'set') then
      vv := v
    elif type(v, 'list') then
      vv := {op(v)}
    else
      vv := {v}
    fi
  fi;
  if not type(f, series) then return false fi;
  x := op(0, f);;
  vv := vv intersect {x};
  if vv = {} then return false fi;
  if has({seq(op(2*i - 1,f), i = 1..nops(f)/2)}, x) then return false fi;
  ff := convert(f, polynom);
  omega := ldegree(ff, x);

  not has({seq(op(2*i,f), i = 1..nops(f)/2)}, x) and
  type(expand(x^(-omega)*ff), polynom(ground_field, x))
end proc:# `type/laurentseries`
#
# formal Puiseux series type
`type/puiseuxseries`:= proc(f, ground_field, v)
  local vv, x;
  option procname;
  if not type(f, 'algebraic') then
    return false
  fi;
  vv := indets(f, name);
  if _npassed = 1 then
    return thisproc(f, anything, vv)
  elif _npassed = 2 then
    return thisproc(f, ground_field, vv)
  elif not _npassed = 3  then
    error "wrong type of arguments"
  else
    if type(v, 'set') then
      vv := v
    elif type(v, 'list') then
      vv := {op(v)}
    else
      vv := {v}
    fi
  fi;
  if type(f, 'function') and op(0, f) = ' dcfun:-PuiseuxSeries'
                       and nops(f) = 2
                       and type(op(1, f), posint)
                       and type(op(2, f), series) then
    x := op([2,0], f);
    vv := vv intersect {x};
    if vv = {} then
      return false
    fi;
    if type(op(2, f), 'laurentseries'(ground_field, x)) then
      return true
    else return false
    fi;
  else
    return false
  fi;
end proc:# `type/puiseuxseries`
#
# pretty printer for Mahlerian hypergeometric functions
`print/MHypergeom` := proc()
  local g, x, b, k;
  g := _passed[1];
  x := _passed[2];
  b := _passed[3];
  k := `tools/genglobal`('_k');
  Product(simplify(eval(1/g, x = x^(b^k)), 'power', 'symbolic'),
          k = 0..infinity)
end proc:# `print/MHypergeom`
#
# pretty printer for Puiseux series
`print/PuiseuxSeries` := proc()
  local q, s, x;
  q := _passed[1];
  s := _passed[2];
  x := op(0, s);
  eval(s, x = x^(1/q))
end proc:# `print/PuiseuxSeries`

# Merge/move if draft.mpl is split into included files.
$include <absprimdec.mm>

end module:
